###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Production options is a utility to return options for projects based on
current LHCb software versions.

This is used by the production API to create production workflows but
also provides lists of options files for test jobs.
"""
import re
from DIRAC import S_OK, gLogger

gLogger = gLogger.getSubLogger("ProductionOptions")


def getModuleOptions(
    applicationName, numberOfEvents, inputDataOptions, extraOptions="", runNumber=0, firstEventNumber=1, jobType=""
):
    """Return the standard options for a Gaudi application project to be used at
    run time by the workflow modules.

    The input data options field is a python list (output of
    getInputDataOptions() below). The runNumber and firstEventNumber only
    apply in the Gauss case and when the job type is not 'user'.
    """
    optionsLines = []
    optionsLines.append("\n\n#////////////////////////////////////////////")
    optionsLines.append("# Dynamically generated options in a gaudirun job\n")
    if applicationName.lower() == "davinci" or applicationName.lower() == "lhcb":
        optionsLines.append("from Gaudi.Configuration import *")
    else:
        optionsLines.append(f"from {applicationName}.Configuration import *")

    if extraOptions:
        for opt in extraOptions.split(";"):
            if opt:
                optionsLines.append(opt)

    if inputDataOptions:
        optionsLines += inputDataOptions

    if applicationName.lower() == "gauss" and not jobType.lower() == "user":
        optionsLines.append('GaussGen = GenInit("GaussGen")')
        optionsLines.append(f"GaussGen.RunNumber = {runNumber}")
        optionsLines.append(f"GaussGen.FirstEventNumber = {firstEventNumber}")

    if numberOfEvents != 0:
        optionsLines.append("ApplicationMgr().EvtMax = %d" % (numberOfEvents))

    finalLines = "\n".join(optionsLines) + "\n"
    return S_OK(finalLines)


def getDataOptions(applicationName, inputDataList, inputDataType, poolXMLCatalogName):
    """Given a list of input data and a specified input data type this function
    will return the correctly formatted EventSelector options for Gaudi
    applications specified by name.

    The options are returned as a python list.
    """
    options = []
    if inputDataList:
        gLogger.info(f"Formatting options for {applicationName}: {len(inputDataList)} LFN(s) of type {inputDataType}")

        inputDataOpt = getEventSelectorInput(inputDataList, inputDataType)["Value"]
        evtSelOpt = f"""EventSelector().Input=[{inputDataOpt}];\n"""
        options.append(evtSelOpt)

    poolOpt = f"""\nFileCatalog().Catalogs= ["xmlcatalog_file:{poolXMLCatalogName}"]\n"""
    options.append(poolOpt)
    return S_OK(options)


def getEventSelectorInput(inputDataList, inputDataType):
    """Returns the correctly formatted event selector options for accessing input
    data using Gaudi applications."""
    inputDataFiles = []
    for lfn in inputDataList:
        lfn = lfn.replace("LFN:", "").replace("lfn:", "")
        if inputDataType == "MDF":
            inputDataFiles.append(f""" "DATAFILE='LFN:{lfn}' SVC='LHCb::MDFSelector'", """)
        elif inputDataType in ("ETC", "SETC", "FETC"):
            cmd = f"COLLECTION='TagCreator/EventTuple' DATAFILE='LFN:{lfn}' "
            cmd += "TYP='POOL_ROOT' SEL='(StrippingGlobal==1)' OPT='READ'"
            inputDataFiles.append(f""" {cmd} """)
        elif inputDataType == "RDST":
            if re.search("rdst$", lfn):
                inputDataFiles.append(f""" "DATAFILE='LFN:{lfn}' TYP='POOL_ROOTTREE' OPT='READ'", """)
            else:
                gLogger.info(f"Ignoring file {lfn} for step with input data type {inputDataType}")
        else:
            inputDataFiles.append(f""" "DATAFILE='LFN:{lfn}' TYP='POOL_ROOTTREE' OPT='READ'", """)

    inputDataOpt = "\n".join(inputDataFiles)[:-2]
    return S_OK(inputDataOpt)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Bookkeeping utilities."""

from DIRAC import gLogger, S_ERROR

from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient


_IGNORE_PARAMETERS = ["ReplicaFlag", "Visible", "MethodName"]

# The following parameters can not used to build the query, it requires at least one more parameter.
_ONE = ["FileType", "ProcessingPass", "EventType", "DataQuality", "ConfigName", "ConfigVersion", "ConditionDescription"]

# Two parameter in the list not enough to build the query.
_TWO = ["ConfigName", "ConfigVersion", "ConditionDescription", "EventType", "ProcessingPass", "FileType", "DataQuality"]


def enoughParams(in_dict):
    """Dirty method to check the query parameters and make sure the queries have
    enough parameters."""
    checkingDict = in_dict.copy()
    if not checkingDict:
        return False

    for param in _IGNORE_PARAMETERS:
        if param in checkingDict:
            checkingDict.pop(param)

    if not checkingDict:
        return False

    if len(checkingDict) == 1:
        if not set(checkingDict) - set(_ONE):
            return False

    if len(checkingDict) == 2:
        if not set(checkingDict) - set(_TWO):
            return False
    return True


def checkEnoughBKArguments(func):
    """The decorator used to check the parameters of a given dictionary
    (BkQuery)."""

    def checkMethodArguments(self, *args, **kwargs):
        """This is used to check the conditions of a given query.

        We assume a dictionary can not be empty and it has more than one
        element, if we do not take into account the replica flag and the
        visibility flag
        """

        if args:
            arguments = args[0]
            if not enoughParams(arguments):
                if isinstance(func, staticmethod):
                    if hasattr(func, "__func__"):
                        funcName = func.__func__.func_name
                    else:
                        # we may do not know the type of the method
                        funcName = repr(func)
                else:
                    funcName = func.func_name

                res = self.getRemoteCredentials()
                userName = res.get("username", "UNKNOWN")
                address = self.email
                subject = f"{funcName} method!"
                body = (
                    "%s user has not provided enough input parameters! \n \
                the input parameters:%s "
                    % (userName, str(arguments))
                )
                NotificationClient().sendMail(address, subject, body, "lhcb-bookkeeping@cern.ch")
                gLogger.error(f"Got you: {userName} ---> {str(arguments)}")
                if self.forceExecution:  # we can force to execute the methods even the user does not
                    # provide enough parameter
                    result = func(self, *args)
                    return result
                else:
                    return S_ERROR(f"Provide more parameters {str(arguments)}")  # TODO: use errno.EINVAL
            else:
                result = func(self, *args)
                return result
        else:
            result = func(self, *args)
            return result

    return checkMethodArguments



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Actual engine for adding a DM transformation, called by dirac-dms-add-
Transformation."""
import os
import json
from collections import defaultdict

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader

from LHCbDIRAC.TransformationSystem.Client.Transformation import Transformation
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import getProcessingPasses, BKQuery
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbDIRAC.TransformationSystem.Utilities.PluginUtilities import getRemovalPlugins, getReplicationPlugins
from LHCbDIRAC.DataManagementSystem.Utilities.FCUtilities import chown
from LHCbDIRAC.DataManagementSystem.Client.DMScript import ProgressBar


def _checkMCReplication(bkPaths):
    """
    Check all MC replication transformations and list those that are obsolete
    """
    # Check only against Idle transformations
    res = TransformationClient().getTransformations(
        {"TransformationGroup": "LHCbMCDSTBroadcastRandom", "Status": ["Idle"]}
    )
    if not res["OK"]:
        gLogger.fatal("Error getting transformations", res["Message"])
        DIRAC.exit(1)
    transPaths = [(dt["TransformationID"], dt["TransformationName"].replace("Replication-", "")) for dt in res["Value"]]
    gLogger.notice(f"Checking {len(transPaths)} transformations against {len(bkPaths)} active BK paths")
    # Only keep BK paths
    transPaths = [trans for trans in transPaths if trans[1].startswith("/MC/")]
    obsoleteTrans = []
    for transID, path in transPaths:
        # Check if name ends with -<number>
        xx = path.split("-")
        if xx[-1].isdigit():
            path = "-".join(xx[:-1])
        if path not in bkPaths:
            obsoleteTrans.append(transID)
    return obsoleteTrans


def getTransformationName(transName, unique):
    """
    Get a transformation name from a base name
    If unique is requested, return None if already exists
    """
    tName = transName
    trial = 0
    transClient = TransformationClient()
    while True:
        # Check if there is already a transformation with that name
        res = transClient.getTransformation(tName)
        if res["OK"]:
            # Transformation already exists
            if unique:
                # If unique is required and the transformation is not in a final status, give up
                if res["Value"]["Status"] not in (
                    "Archived",
                    "Cleaned",
                    "Cleaning",
                    "Deleted",
                    "TransformationCleaned",
                ):
                    tName = None
                    gLogger.notice(
                        "Transformation %s already exists with ID %d, status %s"
                        % (transName, res["Value"]["TransformationID"], res["Value"]["Status"])
                    )
                    break
            trial += 1
            # Check again with new name
            tName = transName + "-" + str(trial)
        else:
            # Transformation doesn't exist, OK
            break
    return tName


def executeAddTransformation(pluginScript):
    """Method for actually adding a DM transformation It takes its options and
    argument values from pluginScript."""
    test = False
    start = False
    force = False
    invisible = False
    fcCheck = True
    unique = False
    bkQuery = None
    depth = None
    userGroup = None
    listProcessingPasses = False
    mcVersionSet = None
    nameOption = None
    checkMCReplication = False
    bodyPlugin = None
    transBody = None

    switches = Script.getUnprocessedSwitches()
    for opt, val in switches:
        if opt in ("s", "Start"):
            start = True
        elif opt == "Test":
            test = True
        elif opt == "Force":
            force = True
        elif opt == "SetInvisible":
            invisible = True
        elif opt == "NoFCCheck":
            fcCheck = False
        elif opt == "Unique":
            unique = True
        elif opt == "Chown":
            userGroup = val.split("/")
            if len(userGroup) != 2 or not userGroup[1].startswith("lhcb_"):
                gLogger.fatal("Wrong user/group")
                DIRAC.exit(2)
        elif opt == "Depth":
            try:
                depth = int(val)
            except ValueError:
                gLogger.fatal("Illegal integer depth:", val)
                DIRAC.exit(2)
        elif opt == "ListProcessingPasses":
            listProcessingPasses = True
        elif opt == "Name":
            nameOption = val
        elif opt == "MCVersion":
            if not mcVersionSet:
                mcVersionSet = {x.lower() for x in val.split(",")}
        elif opt == "CheckMCReplication":
            # Force to check all versions
            mcVersionSet = {"all"}
            checkMCReplication = True
        elif opt == "BodyPlugin":
            bodyPlugin = val
        elif opt == "TransBody":
            transBody = val

    if bodyPlugin and transBody:
        gLogger.fatal("Cannot specify both BodyPlugin and TransBody")
        DIRAC.exit(1)
    if userGroup:
        from DIRAC.Core.Security.ProxyInfo import getProxyInfo

        res = getProxyInfo()
        if not res["OK"]:
            gLogger.fatal("Can't get proxy info", res["Message"])
            DIRAC.exit(1)
        properties = res["Value"].get("groupProperties", [])
        if "FileCatalogManagement" not in properties:
            gLogger.error("You need to use a proxy from a group with FileCatalogManagement")
            DIRAC.exit(5)

    plugin = pluginScript.getOption("Plugin")
    # Default plugin for MCDST replication
    if mcVersionSet and not plugin:
        plugin = "LHCbMCDSTBroadcastRandom"

    if not plugin and not listProcessingPasses:
        gLogger.fatal("ERROR: No plugin supplied...")
        Script.showHelp(exitCode=1)
    prods = pluginScript.getOption("Productions")
    requestID = pluginScript.getOption("RequestID")
    fileType = pluginScript.getOption("FileType")
    pluginParams = pluginScript.getPluginParameters()
    pluginSEParams = pluginScript.getPluginSEParameters()
    requestedLFNs = pluginScript.getOption("LFNs")

    transType = None
    if plugin in getRemovalPlugins():
        transType = "Removal"
    elif plugin in getReplicationPlugins():
        transType = "Replication"
    elif not listProcessingPasses:
        gLogger.notice("This script can only create Removal or Replication plugins")
        gLogger.notice("Replication :", str(getReplicationPlugins()))
        gLogger.notice("Removal     :", str(getRemovalPlugins()))
        gLogger.notice(f"If needed, ask for adding {plugin} to the known list of plugins")
        DIRAC.exit(2)

    bk = BookkeepingClient()
    tr = TransformationClient()

    if plugin in ("DestroyDataset", "DestroyDatasetWhenProcessed") or prods:
        # visible = 'All'
        fcCheck = False

    # If mcVersions are defined, get the processing passes
    if mcVersionSet:
        # We keep only numerical values or "all"
        mcVersions = {mcv for mcv in mcVersionSet if mcv.isdigit() or mcv == "all"}
        if not mcVersions:
            gLogger.fatal("Invalid MC versions", ",".join(sorted(mcVersionSet)))
            DIRAC.exit(2)
        if mcVersions != mcVersionSet:
            gLogger.notice("WARNING: list of MC versions reduced to", ",".join(sorted(mcVersions)))
        # Force transformations names to be unique
        unique = True
        res = ProductionRequestClient().getProductionRequestList(
            0, "RequestID", "DESC", 0, 0, {"RequestState": "Active", "RequestType": "Simulation"}
        )
        if not res["OK"]:
            gLogger.fatal("Error getting production requests", res["Message"])
        prodReq = res["Value"]["Rows"]
        reqDict = defaultdict(set)
        for row in prodReq:
            mcVersion = json.loads(row["Extra"])["mcConfigVersion"]
            # If "all" we take all numerical values, otherwise we take the explicit values
            if ("all" in mcVersions and mcVersion.isdigit()) or mcVersion in mcVersions:
                proPath = row["ProPath"]
                if "/Reco" in proPath:
                    reqDict[mcVersion].add(proPath)
        bkPaths = []
        for mcVersion, proPaths in reqDict.items():
            for proPath in proPaths:
                bkPaths.append(f"/MC/{mcVersion}//{proPath}")
    elif not requestedLFNs:
        # If BK path is given, set it such that one goes through the loop below
        bkPaths = [None]
    else:
        bkPaths = []

    transBKQuery = {}
    bkQueries = []
    bkPaths.sort()
    if checkMCReplication:
        gLogger.notice("List of active BK paths to be kept:")
    for bkPath in sorted(bkPaths):
        if bkPath:
            bkQuery = BKQuery(bkPath)
            if checkMCReplication:
                if "..." in bkPath or "*" in bkPath:
                    gLogger.notice(f"Getting list of BK paths for {bkPath}")
                else:
                    gLogger.notice(f"BK path {bkPath}")
            elif not listProcessingPasses:
                gLogger.notice(f"For BK path: {bkPath}")
        else:
            bkQuery = pluginScript.getBKQuery()
        if not bkQuery and not force:
            gLogger.fatal("No LFNs and no BK query were given...")
            Script.showHelp(exitCode=2)
        if bkQuery:
            processingPass = bkQuery.getProcessingPass()
            if "..." in processingPass or "*" in processingPass:
                bkPath = bkQuery.getPath().replace("RealData", "Real Data")
                if listProcessingPasses:
                    gLogger.notice("List of processing passes for BK path", bkPath)
                processingPasses = getProcessingPasses(bkQuery, depth=depth)
                # Skip AnaProd productions for MC unless explicitly requested
                if mcVersionSet and processingPasses:
                    processingPasses = [pp for pp in processingPasses if "AnaProd" not in pp]
                if processingPasses:
                    if not checkMCReplication:
                        if not listProcessingPasses:
                            gLogger.notice(
                                "Transformations will be launched for the following list of processing passes:"
                            )
                        gLogger.notice("\t" + "\n\t".join(processingPasses))
                else:
                    gLogger.notice("No processing passes matching the BK path")
                    continue
                # Create a list of BK queries, taking into account visibility and if needed excluded file types
                exceptTypes = bkQuery.getExceptFileTypes()
                visible = bkQuery.isVisible()
                for pp in processingPasses:
                    query = BKQuery(bkPath.replace(processingPass, pp), visible=visible)
                    query.setExceptFileTypes(exceptTypes)
                    bkQueries.append(query)
            else:
                bkQueries.append(bkQuery)
        if bkPath and not checkMCReplication:
            gLogger.notice("====================================")

    if listProcessingPasses:
        DIRAC.exit(0)

    # Check all MC replication transformations and list obsolete ones
    if checkMCReplication:
        bkPaths = [bkQuery.getPath() for bkQuery in bkQueries]
        obsoleteTrans = _checkMCReplication(bkPaths)
        gLogger.notice("====================================")
        # Print list of BK paths
        gLogger.notice(
            f"List of {len(obsoleteTrans)} transformations to set Completed:",
            "\n" + ",".join(str(transID) for transID in obsoleteTrans),
        )
        DIRAC.exit(0)

    reqID = pluginScript.getRequestID()
    if not requestID and reqID:
        requestID = reqID

    transGroup = plugin
    # If no BK queries are given, set to None to go once in the loop
    if not bkQueries:
        bkQueries = [None]

    for bkQuery in bkQueries:
        if bkQuery != bkQueries[0]:
            gLogger.notice("**************************************")
        # Create the transformation
        transformation = Transformation()
        transformation.setType(transType)
        transName = transType

        # In case there is a loop on processing passes
        if bkQuery:
            transBKQuery = bkQuery.getQueryDict()
        if requestedLFNs:
            longName = transGroup + f" for {len(requestedLFNs)} LFNs"
            transName += "-LFNs"
        elif prods:
            if not fileType:
                fileType = ["All"]
            prodsStr = ",".join(str(p) for p in prods)
            fileStr = ",".join(fileType)
            longName = transGroup + " of " + fileStr + f" for productions {prodsStr} "
            if len(prods) > 5:
                prodsStr = f"{len(prods)}-productions"
            if len(fileStr) > 30:
                fileStr = f"{len(fileType)}-fileTypes"
            transName += "-" + fileStr + "-" + prodsStr
        elif transBKQuery and "FileType" in transBKQuery and "BKPath" not in pluginScript.getOptions():
            if isinstance(transBKQuery["FileType"], list):
                strQuery = ",".join(transBKQuery["FileType"])
            else:
                strQuery = str(transBKQuery["FileType"])
            longName = transGroup + " for fileType " + strQuery
            transName += "-" + str(transBKQuery["FileType"])
        elif bkQuery:
            queryPath = bkQuery.getPath()
            longName = transGroup + " for BKQuery " + queryPath
            transName += "-" + queryPath
        else:
            transName = ""

        dqFlag = transBKQuery.get("DataQuality", [])
        if "BAD" in dqFlag:
            dqFlag = f" (DQ: {','.join(dqFlag)})"
            transName += dqFlag
            longName += dqFlag

        if requestID:
            transName += f"-Request{requestID}"
        # If a name is given in the options, use it
        if nameOption:
            transName = nameOption
            longName = transGroup + " - " + transName
        if not transName:
            gLogger.fatal("Didn't manage to find a name for this transformation, check options")
            DIRAC.exit(1)
        # Find a name for this transformation (transName remains the base name)
        tName = getTransformationName(transName, unique)
        # If needed, skip this BK query
        if tName is None:
            continue

        transformation.setTransformationName(tName)
        transformation.setTransformationGroup(transGroup)
        transformation.setDescription(longName[:255])
        transformation.setLongDescription(longName)
        transformation.setType(transType)

        # Rename plugin
        if plugin == "DestroyDatasetWhenProcessed":
            plugin = "DeleteReplicasWhenProcessed"
            # Set the polling period to 0 if not defined
            pluginParams.setdefault("Period", 0)

        # If we have a body plugin, load it
        if bodyPlugin:
            objLoader = ObjectLoader()
            _class = objLoader.loadObject(f"TransformationSystem.Client.BodyPlugin.{bodyPlugin}", bodyPlugin)

            if not _class["OK"]:
                raise Exception(_class["Message"])
            # TODO: handle the case of a body plugin with arguments
            transBody = _class["Value"]()
        else:
            # If we don't have a body, define it for Removal transformations
            if not transBody and transType == "Removal":
                if plugin == "DestroyDataset":
                    transBody = "removal;RemoveFile"
                elif plugin == "DestroyDatasetWhenProcessed":
                    transBody = "removal;RemoveFile"
                else:
                    transBody = "removal;RemoveReplica"

        if transBody:
            transformation.setBody(transBody)

        if pluginSEParams:
            for key, val in pluginSEParams.items():
                res = transformation.setSEParam(key, val)
                if not res["OK"]:
                    gLogger.error("Error setting SE parameter", res["Message"])
                    DIRAC.exit(1)
        if pluginParams:
            for key, val in pluginParams.items():
                res = transformation.setAdditionalParam(key, val)
                if not res["OK"]:
                    gLogger.error("Error setting additional parameter", res["Message"])
                    DIRAC.exit(1)

        transformation.setPlugin(plugin)

        if test:
            gLogger.notice("Transformation type:", transType)
            gLogger.notice("Transformation Name:", transName)
            gLogger.notice("Transformation group:", transGroup)
            gLogger.notice("Long description:", longName)
            gLogger.notice("Transformation body:", transBody)
            if transBKQuery:
                gLogger.notice("BK Query:", transBKQuery)
            elif requestedLFNs:
                gLogger.notice(f"List of {len(requestedLFNs)} LFNs")
            else:
                # Should not happen here, but who knows ;-)
                gLogger.error("No BK query provided...")
                Script.showHelp(exitCode=1)

        if force:
            lfns = []
        elif transBKQuery:
            bkPath = bkQuery.getPath()
            title = "Executing BK query"
            if bkPath:
                title += " for " + bkPath
            else:
                title += ": " + str(bkQuery)
            progressBar = ProgressBar(1, title=title)
            lfns = bkQuery.getLFNs(
                printSEUsage=(transType == "Removal" and not pluginScript.getOption("Runs")), printOutput=test
            )
            progressBar.endLoop(message=(f"found {len(lfns)} files" if lfns else "no files found"))
        else:
            lfns = requestedLFNs
        nfiles = len(lfns)

        if test:
            gLogger.notice("Plugin:", plugin)
            gLogger.notice("Parameters:", pluginParams)
            gLogger.notice("RequestID:", requestID)
            continue

        if not force and nfiles == 0:
            gLogger.notice("No files, but if you anyway want to submit the transformation, use option --Force")
            continue

        if userGroup:
            directories = {os.path.dirname(lfn) for lfn in lfns}
            res = chown(directories, user=userGroup[0], group=userGroup[1])
            if not res["OK"]:
                gLogger.fatal("Error changing ownership", res["Message"])
                DIRAC.exit(3)
            gLogger.notice("Successfully changed owner/group for %d directories" % res["Value"])

        # If the transformation is a removal transformation,
        #  check all files are in the FC. If not, remove their replica flag
        if fcCheck and transType == "Removal":
            from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

            fc = FileCatalog()
            success = 0
            missingLFNs = set()
            chunkSize = 1000
            progressBar = ProgressBar(len(lfns), title=f"Checking {len(lfns)} files in FC", chunk=chunkSize)
            for lfnChunk in breakListIntoChunks(lfns, chunkSize):
                progressBar.loop()
                res = fc.exists(lfnChunk)
                if res["OK"]:
                    success += len(res["Value"]["Successful"])
                    missingLFNs |= set(res["Value"]["Failed"])
                else:
                    gLogger.fatal("\nError checking files in the FC", res["Message"])
                    DIRAC.exit(2)
            progressBar.endLoop(message="all found" if not missingLFNs else None)
            if missingLFNs:
                gLogger.notice(
                    "%d are in the FC, %d are not. Attempting to remove GotReplica" % (success, len(missingLFNs))
                )
                res = bk.removeFiles(list(missingLFNs))
                if res["OK"]:
                    gLogger.notice("Replica flag successfully removed in BK")
                else:
                    gLogger.fatal("Error removing BK flag", res["Message"])
                    DIRAC.exit(2)

        # Prepare the transformation
        if transBKQuery:
            # !!!! Hack in order to remain compatible with hte BKQuery table...
            for transKey, bkKey in (
                ("ProductionID", "Production"),
                ("RunNumbers", "RunNumber"),
                ("DataQualityFlag", "DataQuality"),
            ):
                if bkKey in transBKQuery:
                    transBKQuery[transKey] = transBKQuery.pop(bkKey)
            transformation.setBkQuery(transBKQuery)

        # If the transformation uses the RemoveDatasetFromDisk plugin, set the files invisible in the BK...
        # Try and let them visible such that users can see they are archived...
        # It was:
        # setInvisiblePlugins = ("RemoveDatasetFromDisk", )
        setInvisiblePlugins = tuple()
        if invisible or plugin in setInvisiblePlugins:
            chunkSize = 1000
            progressBar = ProgressBar(len(lfns), title=f"Setting {len(lfns)} files invisible", chunk=chunkSize)
            okFiles = 0
            for lfnChunk in breakListIntoChunks(lfns, chunkSize):
                progressBar.loop()
                res = bk.setFilesInvisible(lfnChunk)
                if res["OK"]:
                    okFiles += len(lfnChunk)
            if okFiles == len(lfns):
                msg = "all files successfully set invisible in BK"
            else:
                msg = "%d files successfully set invisible in BK" % okFiles
            progressBar.endLoop(message=msg)
            if res["OK"]:
                if transBKQuery:
                    savedVisi = transBKQuery.get("Visible")
                    transBKQuery["Visible"] = "All"
                    transformation.setBkQuery(transBKQuery.copy())
                    if savedVisi:
                        transBKQuery["Visible"] = savedVisi
                    else:
                        transBKQuery.pop("Visible")
            else:
                gLogger.error("Failed to set files invisible: ", res["Message"])

        errMsg = ""
        while True:
            res = transformation.addTransformation()
            if not res["OK"]:
                errMsg = "Couldn't create transformation"
                break
            res = transformation.getTransformationID()
            if res["OK"]:
                transID = res["Value"]
            else:
                errMsg = "Error getting transformationID"
                break
            # If some LFNs must be added, do it now
            if requestedLFNs:
                from LHCbDIRAC.TransformationSystem.Utilities.PluginUtilities import addFilesToTransformation

                res = addFilesToTransformation(transID, requestedLFNs, addRunInfo=True)
                if not res["OK"]:
                    errMsg = "Could not add files to transformation"
                    break
                gLogger.notice(f"{len(res['Value']['Successful'])} files successfully added to transformation")
                if res["Value"]["Failed"]:
                    errors = defaultdict(int)
                    for error in res["Value"]["Failed"].values():
                        errors[error] += 1
                    for error, count in errors.items():
                        gLogger.always("Failed to add %d files to transformation %d:" % (count, transID), error)
            if requestID:
                transformation.setTransformationFamily(requestID)
            if start:
                transformation.setStatus("Active")
                transformation.setAgentType("Automatic")
            gLogger.notice(f"Name: {transName}")
            gLogger.notice("Transformation body:", transBody)
            gLogger.notice("Plugin:", plugin)
            if pluginParams:
                gLogger.notice("Additional parameters:", pluginParams)
            if requestID:
                gLogger.notice("RequestID:", requestID)
            break
        if errMsg:
            gLogger.notice(errMsg, res["Message"])

    DIRAC.exit(0)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Insert new file types in the Bookkeeping."""


import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage(__doc__ + "\n".join(["Usage:", f"  {Script.scriptName} [option|cfgfile]"]))
    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    bk = BookkeepingClient()

    exitCode = 0

    ftype = input("FileType: ")
    desc = input("Description: ")
    version = input("File type version: ")
    gLogger.notice("Do you want to add this new file type? (yes or no)")
    value = input("Choice:")
    choice = value.lower()
    if choice in ["yes", "y"]:
        res = bk.insertFileTypes(ftype.upper(), desc, version)
        if res["OK"]:
            gLogger.notice("The file types added successfully!")
        else:
            gLogger.error("Error discovered!", res["Message"])
    elif choice in ["no", "n"]:
        gLogger.notice("Aborted!")
    else:
        gLogger.error("Unexpected choice:", value)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""File utilities module (e.g. make GUIDs)"""
import uproot

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.File import makeGuid as DIRACMakeGUID


def getRootFileGUIDs(fileList):
    """Retrieve a list of GUIDs for a list of files."""
    guids = {"Successful": {}, "Failed": {}}
    for fileName in fileList:
        res = getRootFileGUID(fileName)
        if res["OK"]:
            gLogger.verbose("GUID from ROOT", f"{res['Value']}")
            guids["Successful"][fileName] = res["Value"]
        else:
            guids["Failed"][fileName] = res["Message"]
    return S_OK(guids)


def getRootFileGUID(fileName):
    """Function to retrieve a file GUID using uproot."""
    try:
        with uproot.open(fileName) as f:
            branch = f["Refs"]["Params"]
            for item in branch.array():
                if item.startswith("FID="):
                    return S_OK(item.split("=")[1])
        return S_ERROR("GUID not found")
    except Exception as e:
        errorMsg = f"Error extracting GUID: {e}"
        return S_ERROR(errorMsg)


def makeGuid(fileNames):
    """Function to retrieve a file GUID using uproot."""
    if isinstance(fileNames, str):
        fileNames = [fileNames]

    fileGUIDs = {}
    for fileName in fileNames:
        res = getRootFileGUID(fileName)
        if res["OK"]:
            gLogger.verbose("GUID from ROOT", f"{res['Value']}")
            fileGUIDs[fileName] = res["Value"]
        else:
            gLogger.info("Could not obtain GUID from file through Gaudi, using standard DIRAC method")
            fileGUIDs[fileName] = DIRACMakeGUID(fileName)

    return fileGUIDs



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_BKK_DB_OracleBookkeepingDB."""
from unittest import mock

from LHCbDIRAC.BookkeepingSystem.DB.LegacyOracleBookkeepingDB import LegacyOracleBookkeepingDB
from LHCbDIRAC.BookkeepingSystem.DB.OracleDB import OracleDB
from DIRAC.ConfigurationSystem.Client.Config import gConfig


def test_instantiate(monkeypatch):
    """tests that we can instantiate one object of the tested class."""

    module = LegacyOracleBookkeepingDB(dbW=None, dbR=None)
    assert module.__class__.__name__ == "LegacyOracleBookkeepingDB"


def test_buildRunNumbers():
    """It test the method which used to build the conditions when runnumbers is
    a list/number, and end run and start run is a number."""
    client = LegacyOracleBookkeepingDB(dbW=None, dbR=None)
    runnumbers = [1, 3, 4]
    startRunID = None
    endRunID = None
    condition = ""
    tables = ""
    retVal = client._buildRunnumbers(runnumbers, startRunID, endRunID, condition, tables)
    assert retVal["OK"] is True
    outVal = " AND  ( j.runnumber=1 OR  j.runnumber=3 OR  j.runnumber=4 ) "
    assert retVal["Value"] == (outVal, "")

    startRunID = 1
    retVal = client._buildRunnumbers(runnumbers, startRunID, endRunID, condition, tables)
    assert retVal["OK"] is True
    assert retVal["Value"] == (outVal, "")

    startRunID = None
    endRunID = 1
    retVal = client._buildRunnumbers(runnumbers, startRunID, endRunID, condition, tables)
    assert retVal["OK"] is True
    assert retVal["Value"] == (outVal, "")

    startRunID = 1
    endRunID = 2
    runnumbers = [33, 44]
    retVal = client._buildRunnumbers(runnumbers, startRunID, endRunID, condition, tables)
    assert retVal["OK"] is True
    outVal = " AND (j.runnumber>=1 AND j.runnumber<=2 OR  ( j.runnumber=33 OR  j.runnumber=44 )) "
    assert retVal["Value"] == (outVal, "")

    for i in [[], None]:
        runnumbers = i
        startRunID = 1
        endRunID = 2
        retVal = client._buildRunnumbers(runnumbers, startRunID, endRunID, condition, tables)
        assert retVal["OK"] is True
        assert retVal["Value"] == (" AND j.runnumber>=1  AND j.runnumber<=2 ", "")


def test_buildConditions():
    """it test the simulation/data taking condition string creation
    procedure."""
    mock_db = mock.Mock(spec=OracleDB)
    client = LegacyOracleBookkeepingDB(dbW=mock_db, dbR=mock_db)
    condition = ""
    tables = ""
    for i in [[], None, "ALL"]:
        simdesc = i
        daqdesc = "BeamReal"
        mock_db.query.return_value = {"OK": True, "Value": [(1,)]}
        client.dbR_ = mock_db
        retVal = client._buildConditions(simdesc, daqdesc, condition, tables)
        assert retVal["OK"] is True
        assert retVal["Value"] == (
            " AND cont.DAQPERIODID=1 AND cont.DAQPERIODID is not null ",
            " , productionscontainer cont ",
        )


#   ################################################################################
#   def test_buildConfiguration(self):
#     """it test the configuration name and version condition string creation."""
#     client = self.testClass()
#     condition = ''
#     tables = ''
#     configName = ''
#     configVersion = ''
#     retVal = client.__buildConfiguration(configName, configVersion, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual(('', ''), retVal['Value'])

#     configName = 'MC'
#     configVersion = ''
#     retVal = client.__buildConfiguration(configName, configVersion, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual(('', ''), retVal['Value'])

#     configName = 'MC'
#     configVersion = 'MC11a'
#     retVal = client.__buildConfiguration(configName, configVersion, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual(("  and c.ConfigName='MC' and c.ConfigVersion='MC11a' \
# and       j.configurationid=c.configurationid ", " ,configurations c"), retVal['Value'])

#   ################################################################################
#   def test_buildDataquality(self):
#     """it test the data quality condition string creation."""
#     client = self.testClass()
#     condition = ''
#     tables = ''
#     mock_db = mock.Mock(spec=OracleDB)
#     mock_db.query.return_value = {'OK': True, 'Value': [(1,)]}
#     client.dbR_ = mock_db
#     for i in ['ALL', None]:
#       dqFlag = i
#       retVal = client.__buildDataquality(dqFlag, condition, tables)
#       self.assertTrue(retVal['OK'])
#       self.assertTrue(retVal['Value'])
#       self.assertEqual(('', ''), retVal['Value'])

#     dqFlag = 'OK'
#     retVal = client.__buildDataquality(dqFlag, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((' and f.qualityid=1', ''), retVal['Value'])

#     dqFlag = ['OK', 'UNCHECKED', 'BAD']
#     retVal = client.__buildDataquality(dqFlag, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((' and ( f.qualityid=1 or f.qualityid=1 or f.qualityid=1)', ''), retVal['Value'])

#   ################################################################################
#   def test_buildEventType(self):
#     """it test the event type condition string creation."""
#     client = self.testClass()
#     condition = ''
#     tables = ''

#     for i in [0, None, 'ALL']:
#       evt = i
#       retVal = client.__buildEventType(evt, condition, tables)
#       self.assertTrue(retVal['OK'])
#       self.assertTrue(retVal['Value'])
#       self.assertEqual(('', ''), retVal['Value'])

#     evt = 1
#     retVal = client.__buildEventType(evt, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((' and f.eventtypeid=1', ''), retVal['Value'])

#     retVal = client.__buildEventType([], condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual(('', ''), retVal['Value'])

#   ################################################################################
#   def test_buildFileTypes(self):
#     """it test the file type condition string creation."""
#     client = self.testClass()
#     condition = ''
#     tables = ''
#     for i in [None, 'ALL']:
#       ft = i
#       retVal = client.__buildFileTypes(ft, condition, tables)
#       self.assertTrue(retVal['OK'])
#       self.assertTrue(retVal['Value'])
#       self.assertEqual(('', ''), retVal['Value'])

#     retVal = client.__buildFileTypes([], condition, tables)
#     self.assertTrue(retVal['Message'])

#   ################################################################################
#   def test_buildProcessingPass(self):
#     """it test the processing pass condition string creation."""
#     client = self.testClass()
#     condition = ''
#     tables = ''
#     mock_db = mock.Mock(spec=OracleDB)
#     mock_db.query.return_value = {'OK': True, 'Value': [(1,)]}
#     client.dbR_ = mock_db

#     for i in [None, 'ALL']:
#       procpass = i
#       retVal = client.__buildProcessingPass(procpass, condition, tables)
#       self.assertTrue(retVal['OK'])
#       self.assertTrue(retVal['Value'])
#       self.assertEqual(('', ''), retVal['Value'])

#     procpass = '/Sim08/Reco01'
#     retVal = client.__buildProcessingPass(procpass, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((' and j.production=prod.production\
#                       and prod.processingid in (1)', ',productionscontainer prod'), retVal['Value'])

#     procpass = '/Sim08/Reco01'
#     tables = ',productionscontainer prod'
#     retVal = client.__buildProcessingPass(procpass, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((' and j.production=prod.production\
#                       and prod.processingid in (1)', ',productionscontainer prod'), retVal['Value'])

#     mock_db.query.return_value = {'OK': True, 'Value': [(1,), (2,)]}
#     retVal = client.__buildProcessingPass(procpass, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((' and j.production=prod.production\
#                       and prod.processingid in (1,2)', ',productionscontainer prod'), retVal['Value'])

#   ################################################################################

#   def test_buildProduction(self):
#     """it test the production condition string creation."""
#     client = self.testClass()
#     condition = ''
#     tables = ''

#     for i in [None, 'ALL', []]:
#       prod = i
#       retVal = client.__buildProduction(prod, condition, tables)
#       self.assertTrue(retVal['OK'])
#       self.assertTrue(retVal['Value'])
#       self.assertEqual(('', ''), retVal['Value'])

#     prod = 1
#     retVal = client.__buildProduction(prod, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((' and j.production=1', ''), retVal['Value'])

#     prod = [1, 2]
#     retVal = client.__buildProduction(prod, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((' and  (  j.production=1 or  j.production=2 )', ''), retVal['Value'])

#   ################################################################################
#   def test_buildReplicaflag(self):
#     """it test the replica flag condition string creation."""
#     client = self.testClass()
#     condition = ''
#     tables = ',production prod'

#     for i in [None, 'ALL', []]:
#       flag = i
#       retVal = client.__buildReplicaflag(flag, condition, tables)
#       self.assertTrue(retVal['OK'])
#       self.assertTrue(retVal['Value'])
#       self.assertEqual(('', tables), retVal['Value'])

#     flag = 'Yes'
#     retVal = client.__buildReplicaflag(flag, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((" and f.gotreplica='Yes' ", tables), retVal['Value'])

#     flag = 'No'
#     retVal = client.__buildReplicaflag(flag, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((" and f.gotreplica='No' ", tables), retVal['Value'])

#   ################################################################################
#   def test_buildStartenddate(self):
#     """it test the start and end date condition string creation."""
#     client = self.testClass()
#     condition = ''
#     tables = ',production prod'

#     for i in [None, 'ALL', []]:
#       sDate = i
#       eDate = i
#       retVal = client.__buildStartenddate(sDate, eDate, condition, tables)
#       self.assertTrue(retVal['OK'])
#       self.assertTrue(retVal['Value'])
#       self.assertEqual(('', tables), retVal['Value'])

#     sDate = 1
#     eDate = 2
#     retVal = client.__buildStartenddate(sDate, eDate, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((" and f.inserttimestamp >= TO_TIMESTAMP ('1','YYYY-MM-DD HH24:MI:SS') \
# and f.inserttimestamp <= TO_TIMESTAMP ('2','YYYY-MM-DD HH24:MI:SS')", tables), retVal['Value'])

#     sDate = None
#     eDate = 2
#     retVal = client.__buildStartenddate(sDate, eDate, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((" and f.inserttimestamp <= TO_TIMESTAMP ('2','YYYY-MM-DD HH24:MI:SS')", tables),
#                      retVal['Value'])

#     sDate = 1
#     eDate = None
#     retVal = client.__buildStartenddate(sDate, eDate, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     condition, tables2 = retVal['Value']
#     self.assertEqual(154, len(condition))  # the current time is returned and can not compared
#     self.assertEqual(tables, tables2)

#   ################################################################################
#   def test_buildTCKS(self):
#     """it test the TCK condition string creation."""
#     client = self.testClass()
#     condition = ''
#     tables = ',production prod'

#     for i in [None, 'ALL']:
#       tcks = i
#       retVal = client.__buildTCKS(tcks, condition, tables)
#       self.assertTrue(retVal['OK'])
#       self.assertTrue(retVal['Value'])
#       self.assertEqual(('', tables), retVal['Value'])

#     tcks = [1]
#     retVal = client.__buildTCKS(tcks, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((" and (j.tck='1' ) ", tables), retVal['Value'])

#     tcks = "1"
#     retVal = client.__buildTCKS(tcks, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((" and j.tck='1'", tables), retVal['Value'])

#     tcks = [1, 2]
#     retVal = client.__buildTCKS(tcks, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual((" and (j.tck='1' or j.tck='2' ) ", tables), retVal['Value'])

#     tcks = []
#     retVal = client.__buildTCKS(tcks, condition, tables)
#     self.assertTrue(retVal['OK'])
#     self.assertTrue(retVal['Value'])
#     self.assertEqual(('', tables), retVal['Value'])



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.ResourceStatusSystem.Policy.TransferQualityPolicy.

TransferQualityPolicy.__bases__:
  DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase.PolicyBase
"""
from DIRAC import S_OK
from DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase import PolicyBase


class TransferQualityPolicy(PolicyBase):
    """The TransferQualityPolicy class is a policy class to check the transfer
    quality.

    Evaluates the TransferQuality results given by the
    DIRACAccounting.TransferQuality command against a certain set of
    thresholds defined in the CS.
    """

    @staticmethod
    def _evaluate(commandResult):
        """Evaluate policy on Data quality.

        :returns:
            {
              'Status':Error|Unknown|Active|Probing|Banned,
              'Reason':'TransferQuality:None'|'TransferQuality:xx%',
            }
        """

        result = {"Status": None, "Reason": None}

        if not commandResult["OK"]:
            result["Status"] = "Error"
            result["Reason"] = commandResult["Message"]
            return S_OK(result)

        commandResult = commandResult["Value"]

        if not commandResult:
            result["Status"] = "Unknown"
            result["Reason"] = "No values to take a decision"
            return S_OK(result)

        if "Name" not in commandResult:
            result["Status"] = "Error"
            result["Reason"] = 'Missing "Name" key'
            return S_OK(result)

        name = commandResult["Name"]

        if "Mean" not in commandResult:
            result["Status"] = "Error"
            result["Reason"] = 'Missing "Mean" key'
            return S_OK(result)

        mean = commandResult["Mean"]

        if mean is None:
            result["Status"] = "Unknown"
            result["Reason"] = "No values to take a decision"
            return S_OK(result)

        result["Reason"] = "TransferQuality: %d -> " % mean

        # FIXME: policyParameters = Configurations.getPolicyParameters()

        policyParameters = {"Transfer_QUALITY_LOW": 60, "Transfer_QUALITY_HIGH": 90}

        if "FAILOVER".lower() in name.lower():
            if mean < policyParameters["Transfer_QUALITY_LOW"]:
                result["Status"] = "Degraded"
                strReason = "Low"
            elif mean < policyParameters["Transfer_QUALITY_HIGH"]:
                result["Status"] = "Active"
                strReason = "Mean"
            else:
                result["Status"] = "Active"
                strReason = "High"

        else:
            if mean < policyParameters["Transfer_QUALITY_LOW"]:
                result["Status"] = "Degraded"
                strReason = "Low"
            elif mean < policyParameters["Transfer_QUALITY_HIGH"]:
                result["Status"] = "Degraded"
                strReason = "Mean"
            else:
                result["Status"] = "Active"
                strReason = "High"

        result["Reason"] = result["Reason"] + strReason
        return S_OK(result)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_RSS_Policy_JobEfficiencyPolicy."""
import unittest

import LHCbDIRAC.ResourceStatusSystem.Policy.JobWebSummaryEfficiencyPolicy as moduleTested

################################################################################


class JobWebSummaryEfficiencyPolicy_TestCase(unittest.TestCase):
    def setUp(self):
        """Setup."""
        self.moduleTested = moduleTested
        self.testClass = self.moduleTested.JobWebSummaryEfficiencyPolicy

    def tearDown(self):
        """Tear down."""

        del self.moduleTested
        del self.testClass


################################################################################


class JobWebSummaryEfficiencyPolicy_Success(JobWebSummaryEfficiencyPolicy_TestCase):
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class."""

        module = self.testClass()
        self.assertEqual("JobWebSummaryEfficiencyPolicy", module.__class__.__name__)

    def test_evaluate(self):
        """tests the method _evaluate."""

        module = self.testClass()

        res = module._evaluate({"OK": False, "Message": "Bo!"})
        self.assertTrue(res["OK"])
        self.assertEqual("Error", res["Value"]["Status"])
        self.assertEqual("Bo!", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": None})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("No values to take a decision", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": []})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("No values to take a decision", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"A": 1}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Error", res["Value"]["Status"])
        self.assertEqual('"Status" key missing', res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"Status": 1}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Error", res["Value"]["Status"])
        self.assertEqual('"Efficiency" key missing', res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"Status": 1, "Efficiency": 1}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Error", res["Value"]["Status"])
        self.assertEqual('Unknown status "1"', res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"Status": "Good", "Efficiency": 1}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Active", res["Value"]["Status"])
        self.assertEqual("Jobs Efficiency: 1 with status Good", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"Status": "Fair", "Efficiency": 1}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Active", res["Value"]["Status"])
        self.assertEqual("Jobs Efficiency: 1 with status Fair", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"Status": "Poor", "Efficiency": 1}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Degraded", res["Value"]["Status"])
        self.assertEqual("Jobs Efficiency: 1 with status Poor", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"Status": "Idle", "Efficiency": 1}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("Jobs Efficiency: 1 with status Idle", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"Status": "Bad", "Efficiency": 1}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Banned", res["Value"]["Status"])
        self.assertEqual("Jobs Efficiency: 1 with status Bad", res["Value"]["Reason"])


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(JobWebSummaryEfficiencyPolicy_TestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(JobWebSummaryEfficiencyPolicy_Success))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)


################################################################################



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Table/object defintions used in the AnalysisProductionsDB"""
from sqlalchemy import (
    ForeignKeyConstraint,
    Integer,
    Column,
    String,
    JSON,
    DateTime,
    func,
    ForeignKey,
    Float,
    UniqueConstraint,
)
from sqlalchemy.orm import declarative_base, relationship, validates
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import expression
from sqlalchemy.types import TypeEngine

Base = declarative_base()
Base.__table_args__ = {
    "mysql_engine": "InnoDB",
    "mysql_charset": "utf8",
}


class utcnow_add_days(expression.FunctionElement):
    """Sqlalchemy function to return a date now() plus 'days_to_add' days.

    Used to set default datetime as NOW() plus some number of days.

    """

    type: TypeEngine = DateTime()
    inherit_cache: bool = True

    def __init__(self, *args, days_to_add, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if not isinstance(days_to_add, int):
            raise TypeError(f"days_to_add must be an int, got {type(days_to_add)}")
        self._days_to_add = days_to_add


@compiles(utcnow_add_days, "mysql")
def mysql_utcnow_add_days(element, compiler, **kw) -> str:
    """Sqlalchemy function for mysql rendering of utcnow_add_days.

    Part of utcnow_add_days.
    """
    return f"DATE_ADD( UTC_TIMESTAMP, INTERVAL {element._days_to_add} DAY)"


@compiles(utcnow_add_days, "sqlite")
def sqlite_utcnow_add_days(element, compiler, **kw) -> str:
    """Sqlalchemy function for sqlite rendering of utcnow_add_days.

    Part of utcnow_add_days.
    """
    return f"DATE(DATETIME('now'), '+{element._days_to_add} days')"


class Request(Base):
    __tablename__ = "ap_requests"

    VALID_STATES = [
        "waiting",
        "active",
        "replicating",
        "ready",
    ]
    request_id = Column(Integer, primary_key=True)
    filetype = Column(String(64), primary_key=True)
    state = Column(String(16), nullable=False, default="waiting")
    progress = Column(Float, nullable=True)
    last_state_update = Column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),  # pylint: disable=not-callable
        onupdate=func.now(),  # pylint: disable=not-callable
    )

    name = Column(String(256), nullable=False)
    version = Column(String(64), nullable=False)
    auto_tags = relationship("AutoTag", back_populates="request", lazy="selectin")
    # TODO: Use the mutable extension and validate this object better?
    extra_info = Column(JSON(), nullable=False, default=lambda: {"transformations": []})

    @validates("name")
    def convert_lower(self, key, value):
        return value.lower()

    def toDict(self):
        result = {
            "name": self.name,
            "version": self.version,
            "request_id": self.request_id,
            "filetype": self.filetype,
            "state": self.state,
            "last_state_update": self.last_state_update,
            "transformations": self.extra_info["transformations"],
        }
        if self.progress is not None:
            result["progress"] = self.progress
        if "jira_task" in self.extra_info:
            result["jira_task"] = self.extra_info["jira_task"]
        if "merge_request" in self.extra_info:
            result["merge_request"] = self.extra_info["merge_request"]
        return result


class AnalysisSample(Request):
    __tablename__ = "ap_analysis_samples"

    sample_id = Column(Integer, primary_key=True)
    wg = Column(String(16), nullable=False)
    analysis = Column(String(256), nullable=False)
    request_id = Column(Integer, nullable=False)
    filetype = Column(String(64), nullable=False)

    __table_args__ = (
        ForeignKeyConstraint(["request_id", "filetype"], ["ap_requests.request_id", "ap_requests.filetype"]),
    )

    owners = relationship("User", back_populates="sample", lazy="selectin")
    tags = relationship("Tag", back_populates="sample", lazy="selectin")
    publications = relationship("Publication", back_populates="sample", lazy="selectin")

    # Allow this table to be temporally versioned
    validity_start = Column(
        DateTime(timezone=False), nullable=False, server_default=func.now()  # pylint: disable=not-callable
    )
    validity_end = Column(DateTime(timezone=False), nullable=True)

    housekeeping_interaction_due = Column(DateTime(timezone=False), server_default=utcnow_add_days(days_to_add=90))

    @validates("wg", "analysis")
    def convert_lower(self, key, value):
        return value.lower()

    def toDict(self):
        result = Request.toDict(self)
        result.update(
            {
                "wg": self.wg,
                "analysis": self.analysis,
                "sample_id": self.sample_id,
                # TODO: Remove
                "owners": [],
                "validity_start": self.validity_start,
                "validity_end": self.validity_end,
                "housekeeping_interaction_due": self.housekeeping_interaction_due,
                "publications": self.publications or [],
            }
        )
        return result


class User(Base):
    __tablename__ = "ap_users"

    id = Column(Integer, primary_key=True)
    username = Column(String(256), nullable=False)
    sample_id = Column(Integer, ForeignKey("ap_analysis_samples.sample_id"), nullable=False)
    sample = relationship("AnalysisSample", back_populates="owners", lazy="selectin")

    # Allow this table to be temporally versioned
    validity_start = Column(
        DateTime(timezone=False), nullable=False, server_default=func.now()  # pylint: disable=not-callable
    )
    validity_end = Column(DateTime(timezone=False), nullable=True)

    @validates("username")
    def convert_lower(self, key, value):
        return value.lower()


class AnalysisOwner(Base):
    __tablename__ = "ap_analyis_owners"

    wg = Column(String(16), primary_key=True)
    analysis = Column(String(256), primary_key=True)
    username = Column(String(256), primary_key=True)

    @validates("username", "wg", "analysis")
    def convert_lower(self, key, value):
        return value.lower()


class Publication(Base):
    __tablename__ = "ap_publications"

    id = Column(Integer, primary_key=True)
    number = Column(String(64), nullable=False)
    sample_id = Column(Integer, ForeignKey("ap_analysis_samples.sample_id"), nullable=False)
    sample = relationship("AnalysisSample", back_populates="publications", lazy="selectin")

    __table_args__ = (UniqueConstraint("number", "sample_id", name="_pubnumber_sample_id_no_duplicates"),)

    @validates("number")
    def convert_upper(self, key, value):
        return value.upper()


class Tag(Base):
    __tablename__ = "ap_tags"

    id = Column(Integer, primary_key=True)
    name = Column(String(64), nullable=False)
    value = Column(String(64), nullable=False)
    sample_id = Column(Integer, ForeignKey("ap_analysis_samples.sample_id"), nullable=False)
    sample = relationship("AnalysisSample", back_populates="tags", lazy="joined")

    # Allow this table to be temporally versioned
    validity_start = Column(
        DateTime(timezone=False), nullable=False, server_default=func.now()  # pylint: disable=not-callable
    )
    validity_end = Column(DateTime(timezone=False), nullable=True)

    @validates("name", "value")
    def convert_lower(self, key, value):
        return value.lower()


class AutoTag(Base):
    __tablename__ = "ap_auto_tags"

    id = Column(Integer, primary_key=True)
    name = Column(String(64), nullable=False)
    value = Column(String(64), nullable=False)
    # composite foreign keys
    request_id = Column(Integer, nullable=False)
    filetype = Column(String(64), nullable=False)
    request = relationship("Request", back_populates="auto_tags", lazy="joined", enable_typechecks=False)

    __table_args__ = (
        ForeignKeyConstraint(["request_id", "filetype"], ["ap_requests.request_id", "ap_requests.filetype"]),
    )

    @validates("name", "value")
    def convert_lower(self, key, value):
        return value.lower()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""The input data resolution module is a VO-specific plugin that allows to
define VO input data policy in a simple way using existing utilities in DIRAC
or extension code supplied by the VO.

The arguments dictionary from the Job Wrapper includes the file
catalogue result and in principle has all necessary information to
resolve input data for applications.
"""
from DIRAC import S_OK
from DIRAC.WorkloadManagementSystem.Client.PoolXMLSlice import PoolXMLSlice
from DIRAC.WorkloadManagementSystem.Client.InputDataResolution import InputDataResolution as DIRACInputDataResolution

COMPONENT_NAME = "LHCbInputDataResolution"
CREATE_CATALOG = False


class InputDataResolution(DIRACInputDataResolution):
    """Define the Input Data Policy."""

    def __init__(self, argumentsDict, bkkClient=None):
        """Standard constructor."""
        super().__init__(argumentsDict)

        if not bkkClient:
            from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

            self.bkkClient = BookkeepingClient()
        else:
            self.bkkClient = bkkClient

    #############################################################################

    def execute(self):
        """Given the arguments from the Job Wrapper, this function calls existing
        utilities in DIRAC to resolve input data according to LHCb VO policy."""
        result = super().execute()
        if not result["OK"] or not result["Value"].get("Successful", {}):
            return result
        resolvedData = result["Value"]["Successful"]

        resolvedData = self._addPfnType(resolvedData)
        if not resolvedData["OK"]:
            return resolvedData
        resolvedData = resolvedData["Value"]

        # TODO: Below is temporary behaviour to prepend mdf: to resolved TURL(s) for case when not a ROOT file
        # This instructs the Gaudi applications to use root to access different file types e.g. for MDF.
        # In the longer term this should be derived from file catalog metadata information.
        # 24/08/2010 - updated hack to use "mdf:" after udpates from Markus
        for lfn, mdataList in resolvedData.items():
            if not isinstance(mdataList, list):
                mdataList = [mdataList]
            for mdata in mdataList:
                if mdata["pfntype"] == "MDF":
                    mdata["turl"] = "mdf:" + mdata["turl"]
                    self.log.info(f"Prepending mdf: to TURL for {lfn}")

        catalogName = self.arguments["Configuration"].get("CatalogName", "pool_xml_catalog.xml")

        self.log.verbose(f"Catalog name will be: {catalogName}")
        appCatalog = PoolXMLSlice(catalogName)
        check = appCatalog.execute(resolvedData)
        if not check["OK"]:
            return check
        return result

    #############################################################################

    def _addPfnType(self, resolvedData):
        """Add the pfn type to the lfn list in input."""

        typeVersions = self.bkkClient.getFileTypeVersion(list(resolvedData))
        if not typeVersions["OK"]:
            return typeVersions
        typeVersions = typeVersions["Value"]

        for lfn, mdataList in resolvedData.items():
            if not isinstance(mdataList, list):
                mdataList = [mdataList]
            if lfn not in typeVersions:
                self.log.warn(f"The file {lfn} do not exist in the BKK, assuming ROOT, unless it is a RAW (MDF)")
                if lfn.split(".")[-1].lower() == "raw":
                    lfnType = "MDF"
                else:
                    lfnType = "ROOT"
            else:
                self.log.verbose(f"Adding PFN file type {typeVersions[lfn]} for {lfn}")
                lfnType = typeVersions[lfn]

            for mdata in mdataList:
                mdata["pfntype"] = lfnType

        return S_OK(resolvedData)

    #############################################################################


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



###############################################################################
# (c) Copyright 2023 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from sqlalchemy import JSON
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import expression


class json_group_array(expression.FunctionElement):
    """Aggregate function that returns a JSON array comprised of all values in the aggregation.

    Provides compatibility between different database engines for performing JSON aggregation.
    See ``sqlite``'s ``json_group_array`` for details.
    """

    type = JSON()
    inherit_cache = True


@compiles(json_group_array, "mysql")
def mysql_json_group_array(element, compiler, **kw):
    if len(element.clauses) != 1:
        raise ValueError("json_group_array requires 1 argument")
    return f"json_arrayagg({compiler.process(element.clauses, **kw)})"


@compiles(json_group_array, "sqlite")
def sqlite_json_group_array(element, compiler, **kw):
    if len(element.clauses) != 1:
        raise ValueError("json_group_array requires 1 arguments")
    return f"json_group_array({compiler.process(element.clauses, **kw)})"


class json_group_object(expression.FunctionElement):
    """Aggregate function that returns a JSON object comprised of all name/value pairs in the aggregation.

    Provides compatibility between different database engines for performing JSON aggregation.
    See ``sqlite``'s ``json_group_object`` for details.
    """

    type = JSON()
    inherit_cache = True


@compiles(json_group_object, "mysql")
def mysql_json_group_object(element, compiler, **kw):
    if len(element.clauses) != 2:
        raise ValueError("json_group_object requires 2 arguments")
    return f"json_objectagg({compiler.process(element.clauses, **kw)})"


@compiles(json_group_object, "sqlite")
def sqlite_json_group_object(element, compiler, **kw):
    if len(element.clauses) != 2:
        raise ValueError("json_group_object requires 2 arguments")
    return f"json_group_object({compiler.process(element.clauses, **kw)})"



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Create a pool xml catalog slice for the specified LFNs."""
import ast
import os
import time
import shlex
import tempfile

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.Core.Utilities.Subprocess import systemCall


def __getLfnsFromFile(optFiles, gaudiVerbose):
    with tempfile.NamedTemporaryFile(mode="rt", suffix=".py") as tmpFile:
        runOpts = ""
        for opt in optFiles:
            if not os.path.exists(opt):
                gLogger.always("File not found: ", opt)
                DIRAC.exit(1)
            runOpts += opt + " "

        gaudiRun = f"gaudirun.py -n -o {tmpFile.name} {runOpts}"
        gLogger.info("Extract list of input files from", optFiles)

        gLogger.info("lb-run LHCb for getting environment")
        command = "lb-run --siteroot=/cvmfs/lhcb.cern.ch/lib LHCb/latest " + gaudiRun

        res = systemCall(timeout=0, cmdSeq=shlex.split(command))
        if not res["OK"]:
            gLogger.always("Error when parsing options files", res["Message"])
            DIRAC.exit(1)
        if gaudiVerbose:
            print(res["Value"][1])

        result = tmpFile.read()
        optDict = ast.literal_eval(result)

    appInput = optDict.get("EventSelector", {}).get("Input")
    if not appInput:
        gLogger.always("Options file do not set EventSelector().Input")
        DIRAC.exit(1)

    return [inp.split()[0].split("'")[1].replace("LFN:", "") for inp in appInput]


def execute():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    catalog = "pool_xml_catalog.xml"
    depth = 1
    optFiles = []
    newOptFile = ""
    ignore = False
    verbose = False
    gaudiVerbose = False

    dmScript = DMScript()
    dmScript.registerSiteSwitches()
    dmScript.registerFileSwitches()
    Script.registerSwitch("", "Options=", "   List of option files to consider")
    Script.registerSwitch("", "NewOptions=", "   Name of a new options file to be generated with LFNs (default: none)")
    Script.registerSwitch("", "Catalog=", f"   Catalog name (default: {catalog}")
    Script.registerSwitch("", "Depth=", "   Depth for ancestor consideration (default: none)")
    Script.registerSwitch("v", "Verbose", "   Verbose gLogger.always(out")
    Script.registerSwitch("", "Ignore", "   Ignore missing files")
    Script.registerSwitch("", "GaudiVerbose", "   Set Gaudi verbose when parsing option files")

    Script.setUsageMessage(__doc__ + "\n".join(["Usage:", f"  {Script.scriptName} [option|cfgfile] ... "]))
    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers

    try:
        site = DMSHelpers().getShortSiteNames()["CERN"]
    except (AttributeError, KeyError):
        site = "LCG.CERN.cern"

    t0 = time.time()
    switches = Script.getUnprocessedSwitches()
    for o, a in switches:
        if o in ("Options"):
            optFiles += a.split(",")
        elif o in ("Catalog"):
            catalog = a
            ext = os.path.splitext(catalog)
            if len(ext[1]) == 0 and "/dev/" not in catalog:
                catalog += os.path.extsep + "xml"
        elif o in ("Depth"):
            try:
                depth = int(a)
            except Exception:
                gLogger.fatal("Invalid depth, must be integer")
                DIRAC.exit(1)
        elif o in ("v", "Verbose"):
            verbose = True
            gLogger.setLevel("Info")
        elif o in ("NewOptions"):
            newOptFile = a
            ext = os.path.splitext(newOptFile)
            if len(ext[1]) == 0:
                newOptFile += os.path.extsep + "py"
        elif o in ("Ignore"):
            ignore = True
        elif o in ("GaudiVerbose"):
            gaudiVerbose = True

    if depth <= 0:
        gLogger.fatal("Invalid ancestor depth, should be >= 1")
        Script.showHelp(exitCode=1)

    if newOptFile and len(optFiles) != 1:
        if optFiles:
            gLogger.fatal("Generating an options file is only valid with a _single_ file as input")
        else:
            gLogger.fatal("Generating an options file is only valid with an option file as input")
        Script.showHelp()
        DIRAC.exit(1)
    if newOptFile:
        if depth > 1:
            gLogger.warn("New options file required, depth ignored...")
        if ignore:
            gLogger.warn("New options file required, cannot ignore missing files")
            ignore = False
        depth = 1
        if os.path.realpath(newOptFile) == os.path.realpath(optFiles[0]):
            gLogger.fatal(
                "Cannot write a new option file overwriting the original one, select a name different from", newOptFile
            )
            Script.showHelp(exitCode=1)

    sites = dmScript.getOption("Sites", [])
    if sites:
        if len(sites) != 1:
            gLogger.fatal("Provide at most one site...")
            Script.showHelp(exitCode=1)
        site = sites[0]

    if optFiles:
        lfnList = __getLfnsFromFile(optFiles, gaudiVerbose)
    else:
        lfnList = dmScript.getOption("LFNs", [])
    if not lfnList:
        gLogger.fatal("No option files and no list of files given")
        Script.showHelp(exitCode=1)

    rc = 0
    savedLevel = gLogger.getLevel()
    try:
        if depth > 1:
            from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

            bk = BookkeepingClient()
            result = bk.getFileAncestors(lfnList, depth)
            if not result["OK"]:
                gLogger.fatal("Error getting ancestor files...")
                DIRAC.exit(1)
            lfnList = [
                anc["FileName"] for ancestors in result["Value"]["Successful"].values() for anc in ancestors
            ] + list(result["Value"]["Successful"])

        from DIRAC.Interfaces.API.Dirac import Dirac

        if not verbose:
            gLogger.setLevel("Exception")
        if newOptFile:
            catalog = "/dev/null"
        gLogger.info("List of LFNs:", lfnList)
        gLogger.info("Site:", site)
        gLogger.info("Catalog file:", catalog)
        result = Dirac().getInputDataCatalog(lfnList, site, catalog, ignoreMissing=ignore)
        if result["OK"]:
            result = result["Value"]
            if result["Failed"]:
                gLogger.always(
                    "Only a fraction of the input files are present and available at %s (%d missing)"
                    % (site, len(result["Failed"]))
                )
            if not result["Successful"]:
                gLogger.fatal("... and none was actually available!")
                DIRAC.exit(1)
            if newOptFile:
                lfnToPfn = result["Successful"]
                optLines = open(optFiles[0]).readlines()
                fo = open(newOptFile, "w")
                for opt in optLines:
                    for lfn in lfnList:
                        if lfn in opt:
                            # get the tURL for this lfn
                            if lfn in lfnToPfn:
                                pfn = lfnToPfn[lfn][0]["turl"]
                                opt = opt.replace("LFN:", "").replace(lfn, pfn)
                            else:
                                gLogger.error("No tURL found for ", lfn)
                            break
                    fo.write(opt)
                fo.close()
                gLogger.always(f"New options file {newOptFile} successfully created")
            else:
                gLogger.always(f"POOL XML catalog {catalog} successfully created")
                if os.path.exists(catalog + ".temp"):
                    os.remove(catalog + ".temp")
                if os.path.isfile(catalog):
                    catOption, ext = os.path.splitext(catalog)
                    catOption += os.path.extsep + "py"
                    fo = open(catOption, "w")
                    fo.write("FileCatalog().Catalogs = [ 'xmlcatalog_file:" + catalog + "' ]\n")
                    fo.close()
                    gLogger.always(f"==> You must add {catOption} to your list of options file")
        else:
            gLogger.error("Error getting the list of PFNs:", result["Message"])
            rc = 1
    except Exception as e:
        gLogger.exception("Exception caught while creating catalog or option file:", "", e)
        rc = 1

    gLogger.setLevel(savedLevel)
    gLogger.info(f"Total execution time for {len(lfnList)} files: {time.time() - t0:5.2f} seconds")
    DIRAC.exit(rc)


@Script()
def main():
    execute()


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit tests for Workflow Modules utilities."""
import unittest
import copy
import os

from unittest.mock import MagicMock, patch

import six
import pytest

pytestmark = pytest.mark.skipif(six.PY2, reason="This test only supports Python 3")

from DIRAC import gLogger

from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock
from LHCbDIRAC.Workflow.Modules.mock_Commons import (
    prod_id,
    prod_job_id,
    wms_job_id,
    workflowStatus,
    stepStatus,
    step_id,
    step_number,
    step_commons,
    wf_commons,
)

from LHCbDIRAC.Workflow.Modules.ErrorLogging import ErrorLogging


class ModulesApplicationsTestCase(unittest.TestCase):
    """Base class for the Modules Applications test cases."""

    def setUp(self):
        gLogger.setLevel("DEBUG")
        self.maxDiff = None

    def tearDown(self):
        for fileProd in [
            "prodConf_someApp_123_00000456_123_00000456_321.py",
            "appLog",
            "gaudi_extra_options.py",
            "applicationError.txt",
            "someApp",
            "applicationLog.txt",
        ]:
            try:
                os.remove(fileProd)
            except OSError:
                continue


#############################################################################
# LHCbScript.py
#############################################################################


class LHCbScriptSuccess(ModulesApplicationsTestCase):
    def test_execute(self):
        from LHCbDIRAC.Workflow.Modules.LHCbScript import LHCbScript

        lhcbScript = LHCbScript()

        lhcbScript.jobType = "merge"
        lhcbScript.stepInputData = ["foo", "bar"]

        lhcbScript.production_id = prod_id
        lhcbScript.prod_job_id = prod_job_id
        lhcbScript.jobID = wms_job_id
        lhcbScript.workflowStatus = workflowStatus
        lhcbScript.stepStatus = stepStatus
        lhcbScript.workflow_commons = wf_commons
        lhcbScript.step_commons = step_commons[0]
        lhcbScript.step_number = step_number
        lhcbScript.step_id = step_id
        lhcbScript.executable = "ls"
        lhcbScript.applicationLog = "applicationLog.txt"

        # no errors, no input data
        for wf_cs in copy.deepcopy(wf_commons):
            for s_cs in step_commons:
                lhcbScript.workflow_commons = wf_cs
                lhcbScript.step_commons = s_cs
                lhcbScript._setCommand()
                lhcbScript._executeCommand()


class LHCbScriptFailure(ModulesApplicationsTestCase):
    def test_execute(self):
        from LHCbDIRAC.Workflow.Modules.LHCbScript import LHCbScript

        lhcbScript = LHCbScript()

        lhcbScript.jobType = "merge"
        lhcbScript.stepInputData = ["foo", "bar"]

        lhcbScript.production_id = prod_id
        lhcbScript.prod_job_id = prod_job_id
        lhcbScript.jobID = wms_job_id
        lhcbScript.workflowStatus = workflowStatus
        lhcbScript.stepStatus = stepStatus
        lhcbScript.workflow_commons = wf_commons
        lhcbScript.step_commons = step_commons[0]
        lhcbScript.step_number = step_number
        lhcbScript.step_id = step_id

        # no errors, no input data
        for wf_cs in copy.deepcopy(wf_commons):
            for s_cs in step_commons:
                lhcbScript.workflow_commons = wf_cs
                lhcbScript.step_commons = s_cs
                res = lhcbScript.execute()
                self.assertFalse(res["OK"])


#############################################################################
# ErrorLogging.py
#############################################################################


class ErrorLoggingSuccess(ModulesApplicationsTestCase):
    @patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
    def test_excecute(self, _patch):
        er = ErrorLogging()
        er.jobType = "user"

        # no errors, no input data
        for wf_cs in copy.deepcopy(wf_commons):
            for s_cs in step_commons:
                self.assertTrue(
                    er.execute(
                        prod_id, prod_job_id, wms_job_id, workflowStatus, stepStatus, wf_cs, s_cs, step_number, step_id
                    )["OK"]
                )


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(ModulesApplicationsTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(LHCbScriptSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(LHCbScriptFailure))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ErrorLoggingSuccess))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
########################################################################
"""DIRAC Basic Oracle Class It provides access to the basic Oracle methods in a
multithread-safe mode keeping used connections in a python Queue for further
reuse.

These are the coded methods:

__init__( user, passwd, tns, [maxConnsInQueue=10] )

Initializes the Queue and tries to connect to the DB server,
using the _connect method.
"maxConnsInQueue" defines the size of the Queue of open connections
that are kept for reuse. It also defined the maximum number of open
connections available from the object.
maxConnsInQueue = 0 means unlimited and it is not supported.


_except( methodName, exception, errorMessage )

Helper method for exceptions: the "methodName" and the "errorMessage"
are printed with ERROR level, then the "exception" is printed (with
full description if it is a Oracle Exception) and S_ERROR is returned
with the errorMessage and the exception.


_connect()

Attemps connection to DB and sets the _connected flag to True upon success.
Returns S_OK or S_ERROR.


query( cmd, [conn] )

Executes SQL command "cmd".
Gets a connection from the Queue (or open a new one if none is available),
the used connection is  back into the Queue.
If a connection to the the DB is passed as second argument this connection
is used and is not  in the Queue.
Returns S_OK with fetchall() out in Value or S_ERROR upon failure.


_getConnection()

Gets a connection from the Queue (or open a new one if none is available)
Returns S_OK with connection in Value or S_ERROR
the calling method is responsible for closing this connection once it is no
longer needed.
"""
# FIXME: use Connection Pooling
# https://python-oracledb.readthedocs.io/en/latest/user_guide/connection_handling.html#connection-pooling
# FIXME: tnsEntry is named dsn in python-oracledb

import time
import threading
import queue

import oracledb
from oracledb import STRING as oracledb_STRING  # pylint: disable=no-name-in-module
from oracledb import NUMBER as oracledb_NUMBER  # pylint: disable=no-name-in-module

from DIRAC import gLogger
from DIRAC import S_OK, S_ERROR


gInstancesCount = 0
gModeFixed = False

maxConnectRetry = 100
maxArraysize = 5000  # max allowed


class OracleDB:
    """Basic multithreaded DIRAC Oracle Client Class."""

    def __init__(self, userName, password="", tnsEntry="", confDir="", mode="", maxQueueSize=100, **kwargs):
        """set Oracle connection parameters and try to connect."""
        global gInstancesCount
        gInstancesCount += 1

        self.__initialized = False
        self._connected = False

        if "logger" not in dir(self):
            self.logger = gLogger.getSubLogger("Oracle")

        # let the derived class decide what to do with if is not 1
        self._threadsafe = oracledb.threadsafety
        self.logger.debug(f"thread_safe = {self._threadsafe}")

        self.__checkQueueSize(maxQueueSize)

        self.__connect_kwargs = kwargs | dict(
            user=userName,
            password=password,
            dsn=tnsEntry,
            config_dir=confDir,
        )
        self.__mode = mode
        # Create the connection Queue to reuse connections
        self.__connectionQueue = queue.Queue(maxQueueSize)
        # Create the connection Semaphore to limit total number of open connection
        self.__connectionSemaphore = threading.Semaphore(maxQueueSize)

        self.__initialized = True
        self._connect()

        if not self._connected:
            raise RuntimeError("Can not connect, exiting...")

        self.logger.info("===================== Oracle =====================")
        self.logger.info("User:           " + self.__connect_kwargs.get("user"))
        self.logger.info("TNS:            " + self.__connect_kwargs.get("dsn"))
        self.logger.info("==================================================")

    def __del__(self):
        global gInstancesCount

        while 1 and self.__initialized:
            self.__connectionSemaphore.release()
            try:
                connection = self.__connectionQueue.get_nowait()
                connection.close()
            except queue.Empty:
                # self.logger.debug("No more connection in Queue")
                break

    @staticmethod
    def __checkQueueSize(maxQueueSize):
        """the size of the internal queue is limited."""

        if maxQueueSize <= 0:
            raise Exception("OracleDB.__init__: maxQueueSize must positive")
        try:
            test = maxQueueSize - 1
        except TypeError:
            raise TypeError(f"OracleDB.__init__: wrong type for maxQueueSize {type(maxQueueSize)}")

    def _except(self, methodName, x, err):
        """print Oracle error or exeption return S_ERROR with Exception."""

        try:
            raise x
        except oracledb.Error as e:
            self.logger.error(f"{methodName}: {err}", str(e))
            return S_ERROR(f"{err}: ( {e} )")
        except Exception as x:
            self.logger.error(f"{methodName}: {err}", str(x))
            return S_ERROR(f"{err}: ({x})")

    def _connect(self):
        """open connection to Oracle DB and put Connection into Queue set connected
        flag to True and return S_OK return S_ERROR upon failure."""
        self.logger.debug("_connect:", self._connected)
        if self._connected:
            return S_OK()

        self.logger.debug("_connect: Attempting to access DB", f"by user {self.__connect_kwargs.get('user')}.")
        try:
            self.__newConnection()
            self.logger.debug("_connect: Connected.")
            self._connected = True
            return S_OK()
        except Exception as x:
            return self._except("_connect", x, "Could not connect to DB.")

    def query(self, cmd, conn=False, params=[], kwparams={}, pre_inserts=[]):
        """execute Oracle query command return S_OK structure with fetchall result
        as tuple it returns an empty tuple if no matching rows are found return
        S_ERROR upon error.

        Use of params and kwparams to pass bind variabes is strongly encouraged
        to prevent SQL injection and improve performance. See the python-oracledb
        documentation for more information.

        :param str cmd: the SQL string to be executed
        :param conn: the connection to use, optional
        :param list params: positional bind variables to pass to oracledb.Cursor.execute
        :param dict kwparams: named bind variables to pass to oracledb.Cursor.execute
        :param list pre_inserts: list of tuples with the query and data to be inserted before the main query
        """
        self.logger.debug("query:", f"{cmd!r} {params!r} {kwparams!r}")

        retDict = self.__getConnection(conn=conn)
        if not retDict["OK"]:
            return retDict
        connection = retDict["Value"]

        try:
            cursor = connection.cursor()
            cursor.arraysize = maxArraysize

            for pre_insert_query, pre_insert_data in pre_inserts:
                self.logger.debug(
                    "query: Pre-inserting data into temp table", f"{pre_insert_query!r} {pre_insert_data!r}"
                )
                cursor.executemany(pre_insert_query, pre_insert_data)

            if cursor.execute(cmd, *params, **kwparams):
                res = cursor.fetchall()
            else:
                res = ()

            # Log the result limiting it to just 10 records
            if len(res) < 10:
                self.logger.debug("query: Records returned", res)
            else:
                self.logger.debug(
                    "query: First 10 records returned out of",
                    f"{len(res)}: {res[:10]} ...",
                )

            retDict = S_OK(res)
        except Exception as x:
            self.logger.debug("query:", cmd)
            retDict = self._except("query", x, "Execution failed.")
            self.logger.debug("Start Rollback transaction")
            connection.rollback()
            self.logger.debug("End Rollback transaction")

        try:
            connection.commit()
            cursor.close()
        except Exception:
            pass
        if not conn:
            self.__putConnection(connection)

        return retDict

    def executeStoredProcedure(self, packageName, parameters, output=True, array=None, conn=False):
        """executes a stored procedure."""
        self.logger.debug("_query:", packageName + "(" + str(parameters) + ")")

        retDict = self.__getConnection(conn=conn)
        if not retDict["OK"]:
            return retDict
        connection = retDict["Value"]

        try:
            cursor = connection.cursor()
            result = None
            results = None
            if array:
                fArray = array[0]
                if isinstance(fArray, str):
                    result = cursor.arrayvar(oracledb_STRING, array)
                    parameters += [result]
                elif isinstance(fArray, int):
                    result = cursor.arrayvar(oracledb_NUMBER, array)
                    parameters += [result]
                elif isinstance(fArray, list):
                    for i in array:
                        if isinstance(i, (bool, str, int)):
                            parameters += [i]
                        elif i:
                            if isinstance(i[0], str):
                                result = cursor.arrayvar(oracledb_STRING, i)
                                parameters += [result]
                            elif isinstance(i[0], int):
                                result = cursor.arrayvar(oracledb_NUMBER, i)
                                parameters += [result]
                            else:
                                return S_ERROR("The array type is not supported!!!")
                        else:
                            result = cursor.arrayvar(oracledb_STRING, [], 0)
                            parameters += [result]
                else:
                    return S_ERROR("The array type is not supported!!!")
            if output:
                result = connection.cursor()
                result.arraysize = maxArraysize  # 500x faster!!
                parameters += [result]
                cursor.callproc(packageName, parameters)
                results = result.fetchall()
            else:
                cursor.callproc(packageName, parameters)
            retDict = S_OK(results)
        except Exception as x:
            self.logger.debug("query:", packageName + "(" + str(parameters) + ")")
            retDict = self._except("query", x, "Execution failed.")
            connection.rollback()

        try:
            cursor.close()
        except Exception as ex:
            self._except("__getConnection:", ex, "Failed to close a connection")
        if not conn:
            self.__putConnection(connection)

        return retDict

    def executeStoredFunctions(self, packageName, returnType, parameters=None, conn=False):
        """executes a stored function."""
        if parameters is None:
            parameters = []
        retDict = self.__getConnection(conn=conn)
        if not retDict["OK"]:
            return retDict
        connection = retDict["Value"]
        try:
            cursor = connection.cursor()
            cursor.arraysize = maxArraysize
            result = cursor.callfunc(packageName, returnType, parameters)
            retDict = S_OK(result)
        except Exception as x:
            self.logger.debug(f"_query: {packageName} ({parameters})")
            retDict = self._except("_query", x, "Execution failed.")
            connection.rollback()

        try:
            cursor.close()
        except Exception as ex:
            self._except("__getConnection:", ex, "Failed to close a connection")
        if not conn:
            self.__putConnection(connection)
        return retDict

    def __newConnection(self):
        """Create a New connection and put it in the Queue."""
        global gModeFixed
        self.logger.debug("__newConnection:")
        if not gModeFixed:
            gModeFixed = True
            if self.__mode == "":
                try:
                    connection = oracledb.connect(**self.__connect_kwargs)
                except Exception as ex:
                    self.logger.debug(f"Thin mode has failed: {ex}, we will try thick")
                else:
                    self.logger.debug("Using implicit thin mode")
                    self.__putConnection(connection)
                    return
                self.__mode = "Thick"
            if self.__mode == "Thick":
                oracledb.init_oracle_client(config_dir=self.__connect_kwargs.get("config_dir"))
            self.logger.debug(f'Using {"thin" if oracledb.is_thin_mode() else "thick"} mode')
        connection = oracledb.connect(**self.__connect_kwargs)
        self.__putConnection(connection)

    def __putConnection(self, connection):
        """Put a connection in the Queue, if the queue is full, the connection is
        closed."""
        self.logger.debug("__putConnection:")

        # Release the semaphore first, in case something fails
        self.__connectionSemaphore.release()
        try:
            self.__connectionQueue.put_nowait(connection)
        except queue.Full:
            self.logger.debug("__putConnection: Full Queue")
            try:
                connection.close()
            except Exception as x:
                self._except("__putConnection", x, "Failed to put Connection in Queue")
        except Exception as x:
            self._except("__putConnection", x, "Failed to put Connection in Queue")

    def _getConnection(self):
        """Return a new connection to the DB It uses the private method
        __getConnection."""
        self.logger.debug("_getConnection:")

        retDict = self.__getConnection(trial=0)
        self.__connectionSemaphore.release()
        return retDict

    def __getConnection(self, conn=False, trial=0):
        """Return a new connection to the DB, if conn is provided then just return it.

        then try the Queue, if it is empty add a newConnection to the Queue
        and retry it will retry maxConnectRetry to open a new connection and
        will return an error if it fails.
        """
        self.logger.debug("__getConnection:")

        if conn:
            return S_OK(conn)

        try:
            self.__connectionSemaphore.acquire()
            connection = self.__connectionQueue.get_nowait()
            self.logger.debug("__getConnection: Got a connection from Queue")
            if connection:
                try:
                    # This will try to reconnect if the connection has timeout
                    connection.commit()
                except BaseException:
                    # if the ping fails try with a new connection from the Queue
                    self.__connectionSemaphore.release()
                    return self.__getConnection()
                return S_OK(connection)
        except queue.Empty:
            self.__connectionSemaphore.release()
            self.logger.debug("__getConnection: Empty Queue")
            try:
                if trial == min(100, maxConnectRetry):
                    return S_ERROR(f"Could not get a connection after {maxConnectRetry} retries.")
                try:
                    self.__newConnection()
                    return self.__getConnection()
                except Exception as x:
                    self.logger.debug("__getConnection: Fails to get connection from Queue", x)
                    time.sleep(trial * 5.0)
                    newtrial = trial + 1
                    return self.__getConnection(trial=newtrial)
            except Exception as x:
                return self._except("__getConnection:", x, "Failed to get connection from Queue")
        except Exception as x:
            return self._except("__getConnection:", x, "Failed to get connection from Queue")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.ResourceStatusSystem.Policy.Configurations.

POLICIESMETA_LHCB policies POLICIESMETA
"""

from DIRAC.ResourceStatusSystem.Policy.Configurations import POLICIESMETA

# LHCb Policies

POLICIESMETA_LHCB = {
    "TransferQualitySource": {
        "description": "Transfers from element, quality measure",
        "module": "TransferQualityPolicy",
        "command": ("TransferCommand", "TransferCommand"),
        "args": {"direction": "Source", "metric": "Quality", "hours": 2},
    },
    "TransferQualityDestination": {
        "description": "Transfers to element, quality measure",
        "module": "TransferQualityPolicy",
        "command": ("TransferCommand", "TransferCommand"),
        "args": {"direction": "Destination", "metric": "Quality", "hours": 2},
    },
}

################################################################################
# Old stuff

policies = {
    "DTScheduled": {
        "description": "Ongoing and scheduled down-times",
        "module": "DTPolicy",
        "commandInNewRes": ("GOCDBStatusCommand", "GOCDBStatusCommand"),
        # 'command'         : ( 'GOCDBStatusCommand', 'DTCachedCommand' ),
        "command": ("GOCDBStatusCommand", "GOCDBStatusCommand"),
        "args": {"hours": 12},  # Fix to avoid querying the CS on load time, to be fixed
        "Site_Panel": [{"WebLink": {"CommandIn": ("GOCDBStatusCommand", "DTInfoCachedCommand"), "args": None}}],
        "Resource_Panel": [{"WebLink": {"CommandIn": ("GOCDBStatusCommand", "DTInfoCachedCommand"), "args": None}}],
    },
    "OnStorageElementPropagation": {
        "description": "How the storage element's nodes are behaving in the RSS",
        "module": "DownHillPropagationPolicy",
        "command": ("RSCommand", "MonitoredStatusCommand"),
        "args": ("Resource",),
        "SE_Panel": [{"RSS": "ResOfStorEl"}],
    },
    "TransferQuality": {
        "description": "SE transfer quality",
        "module": "TransferQualityPolicy",
        "commandInNewRes": ("DIRACAccountingCommand", "TransferQualityCommand"),
        "argsNewRes": None,
        "command": ("DIRACAccountingCommand", "TransferQualityFromCachedPlotCommand"),
        "args": ("DataOperation", "TransferQualityByDestSplitted_2"),
        "SE_Panel": [
            {
                "FillChart - Transfer quality in the last 24 hours": {
                    "CommandIn": ("DIRACAccountingCommand", "CachedPlotCommand"),
                    "args": ("DataOperation", "TransferQualityByDestSplitted_24"),
                    "CommandInNewRes": ("DIRACAccountingCommand", "DIRACAccountingCommand"),
                    "argsNewRes": (
                        "DataOperation",
                        "Quality",
                        {"Format": "LastHours", "hours": 24},
                        "Destination",
                        {"OperationType": "putAndRegister"},
                    ),
                }
            },
        ],
    },
}

# Update DIRAC policies with LHCbDIRAC policies
POLICIESMETA.update(POLICIESMETA_LHCB)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve from Bookkeeping the number of Jobs at each Site for a given Production."""
import DIRAC
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... ProdID",
                "Arguments:",
                "  ProdID:   Production ID (mandatory)",
            ]
        )
    )
    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()
    print("args =", args)
    if len(args) < 1:
        Script.showHelp()

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    bk = BookkeepingClient()
    try:
        prod = int(args[0])
    except BaseException:
        Script.showHelp(exitCode=1)

    res = bk.getNbOfJobsBySites(prod)

    if res["OK"]:
        if not res["Value"]:
            print("No jobs for production", prod)
            DIRAC.exit(0)
        sites = {site: num for num, site in res["Value"]}
        shift = 0
        for site in sites:
            shift = max(shift, len(site) + 2)
        print("Site Name".ljust(shift), "Number of jobs")
        for site in sorted(sites):
            print(site.ljust(shift), str(sites[site]))
    else:
        print(f"ERROR getting number of jobs for {str(prod)}:", res["Message"])


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Create production requests from a YAML document"""
import json
from pathlib import Path

import yaml

from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.Core.Utilities.ReturnValues import convertToReturnValue, returnValueOrRaise

from LHCbDIRAC.ProductionManagementSystem.Utilities.Models import parse_obj, ProductionBase


def parseArgs():
    doSubmit = False
    createFiletypes = False
    outputFilename: Path | None = None

    @convertToReturnValue
    def enableSubmit(_):
        nonlocal doSubmit
        doSubmit = True

    @convertToReturnValue
    def enableCreateFiletypes(_):
        nonlocal createFiletypes
        createFiletypes = True

    @convertToReturnValue
    def setOutputFilename(filename):
        nonlocal outputFilename
        outputFilename = Path(filename)

    switches = [
        ("", "submit", "Actually create steps and submit productions", enableSubmit),
        ("", "create-filetypes", "Create missing file types", enableCreateFiletypes),
        ("", "output-json=", "Write the production IDs to a JSON file", setOutputFilename),
    ]
    Script.registerSwitches(switches)
    Script.registerArgument("yaml_path: Path to the YAML file containing productions to submit")
    Script.parseCommandLine(ignoreErrors=False)
    (yaml_path,) = Script.getPositionalArgs()
    return Path(yaml_path), doSubmit, createFiletypes, outputFilename


@Script()
def main():
    yamlPath, doSubmit, createFiletypes, outputFilename = parseArgs()

    productionRequests = [parse_obj(spec) for spec in yaml.safe_load(yamlPath.read_text())]
    productionIDs = submitProductionRequests(productionRequests, dryRun=not doSubmit, createFiletypes=createFiletypes)
    if productionIDs and outputFilename:
        outputFilename.write_text(json.dumps(productionIDs))
    if not doSubmit:
        gLogger.always('This was a dry run! Pass "--submit" to actually submit production requests.')


def submitProductionRequests(
    productionRequests: list[ProductionBase], *, dryRun=True, createFiletypes
) -> list[tuple[int, int | None, dict[str, int | None]]]:
    """Submit a collection of production requests

    :param productionRequests: List of production requests to submit
    :param dryRun: Set to False to actually submit the production requests
    """
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ModelCompatibility import retValToListOfDict

    # Register filetypes
    requiredFileTypes = set()
    for prod in productionRequests:
        for step in prod.steps:
            requiredFileTypes |= {x.type for x in step.input}
            requiredFileTypes |= {x.type for x in step.output}
    knownFileTypes = {x["FileType"] for x in retValToListOfDict(BookkeepingClient().getAvailableFileTypes())}
    if missingFileTypes := requiredFileTypes - knownFileTypes:
        if not createFiletypes:
            raise NotImplementedError(f"Unknown file types that need to be registered: {missingFileTypes!r}")
        if not dryRun:
            for missingFileType in missingFileTypes:
                returnValueOrRaise(BookkeepingClient().insertFileTypes(missingFileType.upper(), "", "1"))

    # Create steps and submit production requests
    productionIDs = []
    for i, prod in enumerate(productionRequests):
        gLogger.always("Considering production", f"{i+1} of {len(productionRequests)}: {prod.name}")
        prod_id, sub_prod_id = _submitProductionRequests(prod, dryRun=dryRun)
        productionIDs.append([i, prod_id, sub_prod_id])
    return productionIDs


def _submitProductionRequests(prod: ProductionBase, *, dryRun=True) -> tuple[int | None, dict[str, int | None]]:
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
    from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
    from LHCbDIRAC.ProductionManagementSystem.Utilities.Models import ProductionStates
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ModelCompatibility import (
        find_step_id,
        make_subprod_legacy_dict,
        production_to_legacy_dict,
        step_to_step_manager_dict,
    )

    prc = ProductionRequestClient()

    for j, step in enumerate(prod.steps, start=1):
        step.id = find_step_id(j, step)
        if step.id is not None:
            gLogger.info(f"Step {j} of {len(prod.steps)}: Found existing step with ID {step.id=}")
            continue

        if step.application.nightly is not None:
            raise ValueError("Nightly builds cannot be used for submitted productions")

        step_info = step_to_step_manager_dict(j, step)
        gLogger.verbose("Running insertStep with", step_info)
        if not dryRun:
            step.id = returnValueOrRaise(BookkeepingClient().insertStep(step_info))
            gLogger.info(f"Step {j} of {len(prod.steps)}: Created step with ID {step.id=}")

    if prod.id is not None:
        raise RuntimeError(f"{prod.id} has already been submitted")
    request_info, sub_productions = production_to_legacy_dict(prod)
    gLogger.verbose(f"Creating production request with", request_info)
    if not dryRun:
        prod.id = returnValueOrRaise(prc.createProductionRequest(request_info))

    sub_prod_ids = {}
    for sub_prod in sub_productions:
        if prod.state != ProductionStates.NEW:
            raise RuntimeError("Can only add sub productions to productions in state 'New'")
        sub_prod_info = make_subprod_legacy_dict(sub_prod, prod.id)
        gLogger.verbose("Creating production sub request with", request_info)
        sub_prod_id = None if dryRun else returnValueOrRaise(prc.createProductionRequest(sub_prod_info))

        event_type = sub_prod["EventType"]
        if event_type in sub_prod_ids:
            raise NotImplementedError(f"Duplicate event type {event_type} in sub productions")
        sub_prod_ids[event_type] = sub_prod_id

    prod.state = ProductionStates.SUBMITTED
    if not dryRun:
        returnValueOrRaise(prc.updateProductionRequest(prod.id, {"RequestState": prod.state.value}))
        gLogger.always(f"Submitted production {prod.id} with sub productions {sub_prod_ids}")
    return prod.id, sub_prod_ids


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set ExtendedDQOK systems for the given run."""
import DIRAC
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.registerSwitch("", "Overwrite", "Overwrite already registered list")
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... Run [Systems]",
                "Arguments:",
                "  Run:      Run number",
                "  Systems:  Comma separated list of system names (unset all when not specified)",
            ]
        )
    )
    Script.parseCommandLine(ignoreErrors=True)
    overwrite = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "Overwrite":
            overwrite = True
    args = Script.getPositionalArgs()

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    bk = BookkeepingClient()

    systems = []
    if len(args) not in {1, 2}:
        Script.showHelp()
    elif len(args) == 2:
        systems = args[1].split(",")

    exitCode = 0
    rnb = int(args[0])
    result = bk.setExtendedDQOK(rnb, overwrite, systems)

    if not result["OK"]:
        if "ORA-20007" in result["Message"]:
            print("ERROR: different list is already registered for the run (use --Overwrite when desired)")
        else:
            print(f"ERROR: {result['Message']}")
        exitCode = 2
    else:
        print("The list was updated")

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Report sisters or cousins (i.e. descendant of a parent or ancestor) for a (list of) LFN(s)"""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerFileSwitches()
    Script.registerSwitch("", "All", "Do not restrict to sisters with replicas")
    Script.registerSwitch("", "Full", "Get full metadata information on sisters")
    level = 1
    Script.registerSwitch(
        "", "Depth=", "Number of ancestor levels (default: %d), 2 would be cousins, 3 grand-cousins etc..." % level
    )
    Script.registerSwitch("", "Production=", "Production to check for sisters (default=same production)")
    Script.registerSwitch("", "AllFileTypes", "Consider also files with a different type")
    Script.setUsageMessage(
        __doc__ + "\n".join(["Usage:", f"  {Script.scriptName} [option|cfgfile] ... [LFN|File] [Level]"])
    )

    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.BookkeepingSystem.Client.ScriptExecutors import executeFileSisters

    executeFileSisters(dmScript, level=level)


if __name__ == "__main__":
    main()



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""This script is used to flag a given files which belongs a certain run.

1. We flag OK or BAD without specifying a processing pass: this flags the RAW and all derived files as OK or BAD
2. If a processing pass is specified (and we allow one PP at a time only):
     2.1 For BAD, only that processing pass (and derived) is flagged BAD.
         RAW is left unchanged unless PP is '/Real Data' which is then similar to 1
     2.2 For OK:
               - if '/Real Data', only the RAW are flagged OK.
                 Derived data are left unchanged. If one wants to flag everything, use 1.
               - else that processing pass (and derived) and the RAW are flagged OK
"""
import os

import DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Base.Script import Script


def checkDQFlag(dqFlag):
    """Make sure the DQ flag is a known one."""
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    res = BookkeepingClient().getAvailableDataQuality()
    if not res["OK"]:
        return res

    if dqFlag in res["Value"]:
        return S_OK()
    return S_ERROR("Data quality flag is not in the db")


def flagFileList(filename, dqFlag):
    """Flag a LFN or a list of LFN contained in a file."""
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    lfns = []

    # Load the list of LFN
    try:
        with open(filename) as f:
            for lfn in f:
                lfns.append(lfn.strip())
    except OSError:
        lfns = filename.split(",")

    # Now flag the LFN
    res = BookkeepingClient().setFileDataQuality(lfns, dqFlag)

    if not res["OK"]:
        return res

    gLogger.notice("The data quality has been set", f"({dqFlag} for {len(lfns)} files)")
    return S_OK()


def flagRun(runNumber, procPass, dqFlag, flagRAW=False):
    """Flag a run given its number, the processing pass and the DQ flag."""
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    res = getProcessingPasses(runNumber, "/Real Data")
    if not res["OK"]:
        return S_ERROR(f"flagRun: {res['Message']}")

    allProcessingPasses = set(res["Value"])

    gLogger.notice("Available processing passes:", str(allProcessingPasses))

    processingPasses = []
    if procPass:  # we will flag a certain processing: for example: /Real Data/Reco09
        for processingPass in allProcessingPasses:
            if processingPass.startswith(procPass):
                processingPasses.append(processingPass)
    else:  # we will flag everything
        processingPasses = allProcessingPasses

    if flagRAW:
        processingPasses.append("/Real Data")

    if not processingPasses:
        return S_ERROR(f"{procPass} is not a valid processing pass.")

        # Flag the processing passes
    for processingPass in processingPasses:
        res = BookkeepingClient().setRunAndProcessingPassDataQuality(
            runNumber,
            processingPass,
            dqFlag,
        )
        if not res["OK"]:
            return S_ERROR("flagRun: processing pass {processingPass}\n error: {res['Message']}")
        gLogger.notice(f"Run {runNumber} Processing Pass {processingPass} flagged {dqFlag}")

    return S_OK()


def getProcessingPasses(runNumber, procPass):
    """Find all known processing passes for the selected configurations."""
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    res = BookkeepingClient().getRunConfigurationsAndDataTakingCondition(int(runNumber))
    if not res["OK"]:
        return res

    bkDict = res["Value"]
    bkDict["RunNumber"] = runNumber

    passes = []
    res = browseBkkPath(bkDict, procPass, passes)
    if not res["OK"]:
        return res
    return S_OK(passes)


def browseBkkPath(bkDict, processingPass, visitedProcessingPass):
    """This method visit the processing passes started from processingPass.

    The visited processing passes are kept in visitedProcessingPass
    """
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    res = BookkeepingClient().getProcessingPass(bkDict, processingPass)
    if not res["OK"]:
        gLogger.error(
            "Cannot load the processing passes for head % in Version %s Data taking condition %s"
            % (processingPass, bkDict["ConfigVersion"], bkDict["ConditionDescription"])
        )
        gLogger.error(res["Message"])
        return res

    records = res["Value"][0]  # 0 contains the processing passes 1 contains the event types
    if "Name" in records["ParameterNames"]:  # this mean we have processing passes
        # this is the name of the processing pass: 'ParameterNames': ['Name']
        index = records["ParameterNames"].index("Name")
        passes = sorted(os.path.join(processingPass, record[index]) for record in records["Records"])
    else:
        passes = []

    if passes:
        for pName in passes:
            visitedProcessingPass.append(pName)
            browseBkkPath(bkDict, pName, visitedProcessingPass)

    return S_OK()


@Script()
def main():
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    Script.registerSwitch("l:", "lfnfile=", "Flag a LFN or list of LFN")
    Script.registerSwitch("r:", "run=", "Flag a run")
    Script.registerSwitch("p:", "processingPass=", "Processing pass for which a run should be flagged")
    Script.registerSwitch("q:", "dataqualityflag=", "Data quality flag")

    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "\nArguments:",
                "  lfnfile (string) : [LFN|filename]",
                "  processingPass (string) : porcessing pass(es)",
                "  runNumber (int): run number to be flagged",
                "  dataqualityflag (string): data quality\n",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=True)

    params = {"lfn": None, "runnumber": None, "dqflag": None, "processingpass": None}

    for switch in Script.getUnprocessedSwitches():
        if switch[0].lower() in ("l", "lfnfile"):
            params["lfn"] = switch[1]
        elif switch[0].lower() in ("r", "run"):
            params["runnumber"] = int(switch[1])
        elif switch[0].lower() in ("q", "dataqualityflag"):
            params["dqflag"] = switch[1]
        elif switch[0].lower() in ("p", "processingPass"):
            params["processingpass"] = switch[1]

    if params["lfn"] is None and params["runnumber"] is None:
        gLogger.fatal("Please specify run number or an lfn!")
        DIRAC.exit(1)
    if params["dqflag"] is None:
        gLogger.fatal("Please specify the data quality flag!")
        DIRAC.exit(1)

    processingPass = None
    if params["processingpass"]:
        if len(params["processingpass"].split(",")) > 1:
            gLogger.fatal("More than one processing pass is given. Please use only one processing pass!")
            DIRAC.exit(1)
        else:
            processingPass = params["processingpass"]
            if not processingPass.startswith("/"):
                processingPass = "/" + processingPass
            processingPass = processingPass.replace(
                "RealData", "Real Data"
            )  # To be consistent with other BK scripts ;-)
            if not processingPass.startswith("/Real Data"):
                processingPass = "/Real Data" + processingPass

    res = checkDQFlag(params["dqflag"])
    if not res["OK"]:
        gLogger.fatal(f"{params['dqflag']} - {res['Message']}")
        DIRAC.exit(1)

    if params["lfn"]:
        res = flagFileList(params["lfn"], params["dqflag"])
        if res["OK"]:
            gLogger.notice("Files are flagged!")
            DIRAC.exit()
        else:
            gLogger.fatal(res["Message"])

    if params["runnumber"]:
        if not processingPass:  # processing pass is not given
            # this flags the RAW and all derived files as the given dq flag
            res = flagRun(runNumber=params["runnumber"], procPass="/", dqFlag=params["dqflag"], flagRAW=True)
            if not res["OK"]:
                gLogger.fatal(res["Message"])
                DIRAC.exit(1)
        else:  # the processing pass is given
            if (
                params["dqflag"] == "BAD"
            ):  # only that processing pass (and derived) is flagged BAD. RAW is left unchanged
                res = flagRun(
                    runNumber=params["runnumber"],
                    procPass=processingPass,
                    dqFlag=params["dqflag"],
                )
                if not res["OK"]:
                    gLogger.fatal(res["Message"])
                    DIRAC.exit(1)
            elif params["dqflag"] == "OK":
                if processingPass == "/Real Data":  # only flag the RAW
                    res = BookkeepingClient().setRunAndProcessingPassDataQuality(
                        params["runnumber"],
                        "/Real Data",
                        params["dqflag"],
                    )
                    if not res["OK"]:
                        gLogger.fatal(res["Message"])
                        DIRAC.exit(1)
                    else:
                        gLogger.notice("%d flagged OK" % params["runnumber"])
                else:
                    res = flagRun(
                        runNumber=params["runnumber"],
                        procPass=processingPass,
                        dqFlag=params["dqflag"],
                        flagRAW=True,
                    )
                    if not res["OK"]:
                        gLogger.fatal(res["Message"])
                        DIRAC.exit(1)

    DIRAC.exit()


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module holding MCStatsClient class."""
from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.Core.Utilities.ReturnValues import DReturnType


@createClient("ProductionManagement/MCStatsElasticDB")
class MCStatsClient(Client):
    """Client for MCStatsElasticDB"""

    def __init__(self, **kwargs):
        """simple constructor."""

        super().__init__(**kwargs)
        self.setServer("ProductionManagement/MCStatsElasticDB")

    def set(self, typeName: str, data: dict) -> DReturnType:
        """set some data in a certain type.

        :params typeName: type name (e.g. 'errors')
        :params data: dictionary of data to insert

        :returns: S_OK/S_ERROR
        """
        return self._getRPC().set(typeName, data)

    def get(self, typeName: str, productionID: int) -> DReturnType:
        """get per production ID

        :params typeName: type name (e.g. 'errors')
        :params productionID: production ID
        """
        return self._getRPC().get(typeName, productionID)

    def remove(self, typeName: str, productionID: int) -> DReturnType:
        """remove data for productionID.

        :params typeName: type name (e.g. 'errors')
        :params productionID: production ID
        """
        return self._getRPC().remove(typeName, productionID)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""The Sprucing template creates workflows for the following use-cases:
  WORKFLOW1: Sprucing
  WORKFLOW2: Sprucing+Merge
"""

import ast

from DIRAC import initialize

initialize()

from DIRAC import gConfig, gLogger, exit as DIRACexit
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequest import ProductionRequest

gLogger = gLogger.getSubLogger("Sprucing_run.py")

pr = ProductionRequest()

stepsList = ["{{p1Step}}"]
stepsList.append("{{p2Step}}")
stepsList.append("{{p3Step}}")
stepsList.append("{{p4Step}}")
stepsList.append("{{p5Step}}")
stepsList.append("{{p6Step}}")
stepsList.append("{{p7Step}}")
stepsList.append("{{p8Step}}")
stepsList.append("{{p9Step}}")
pr.stepsList = stepsList

###########################################
# Configurable and fixed parameters
###########################################

pr.appendName = "{{WorkflowAppendName#GENERAL: Workflow string to append to production name#1}}"

w = "{{w#----->WORKFLOW: choose one below#}}"
w1 = "{{w1#-WORKFLOW1: Sprucing#False}}"
w2 = "{{w2#-WORKFLOW2: Sprucing+Merge#False}}"

validationFlag = "{{validationFlag#GENERAL: Set True for validation prod#False}}"

# workflow params for all productions
pr.startRun = "{{startRun#GENERAL (obligatory): run start, to set the start run#0}}"
pr.endRun = "{{endRun#GENERAL (obligatory): run end, to set the end of the range#0}}"
pr.runsList = "{{runsList#GENERAL: discrete list of run numbers (do not mix with start/endrun)#}}"
pr.derivedProduction = "{{AncestorProd#GENERAL: ancestor prod to be derived#0}}"
eventTimeouts = "{{eventTimeouts#GENERAL: event timeouts as python dict stepNumber:seconds#}}"
if eventTimeouts:
    pr.eventTimeouts = ast.literal_eval(eventTimeouts)

# spruc params
sprucPriority = int("{{priority#PROD-1:sprucing: priority#2}}")
sprucCPU = "{{sprucMaxCPUTime#PROD-1:sprucing: Max CPU time in secs#1000000}}"
sprucPlugin = "{{sprucPluginType#PROD-1:sprucing: plugin name#Sprucing}}"
sprucFilesPerJob = "{{sprucFilesPerJob#PROD-1:sprucing: Group size or number of files per job#2}}"
sprucDataSE = "{{sprucStreamSE#PROD-1:sprucing: output data SE (un-merged streams)#Tier1-Buffer}}"
try:
    sprucDataSESpecial = ast.literal_eval("{{sprucDataSESpecial#PROD-1:sprucing: Special SE (a dictionary Type:SE)#}}")
except SyntaxError:
    sprucDataSESpecial = {}
sprucAncestorDepth = int("{{sprucAncestorDepth#PROD-1: Ancestor Depth#0}}")
sprucCompressionLvl = "{{sprucCompressionLvl#PROD-1: compression level#LOW}}"
sprucOutputVisFlag = "{{sprucOutputVisFlag#PROD-1: Visibility flag of output files#N}}"
try:
    sprucOutputVisFlagSpecial = ast.literal_eval(
        "{{sprucOutputVisFlagSpecial#PROD-1: Special Visibility flag of output files (dict FType:Y|N)#}}"
    )
except SyntaxError:
    sprucOutputVisFlagSpecial = {}

# merging params
mergingPriority = int("{{MergePriority#PROD-2:Merging: priority#8}}")
mergingCPU = "{{MergeMaxCPUTime#PROD-2:Merging: Max CPU time in secs#300000}}"
mergingPlugin = "{{MergePlugin#PROD-2:Merging: plugin#ByRunFileTypeSizeWithFlush}}"
mergingGroupSize = "{{MergeFileSize#PROD-2:Merging: Size (in GB) of the merged files#5}}"
mergingDataSE = "{{MergeStreamSE#PROD-2:Merging: output data SE (merged streams)#Tier1-DST}}"
try:
    mergingDataSESpecial = ast.literal_eval(
        "{{MergingDataSESpecial#PROD-2:Merging: Special SE (a dictionary Type:SE)#}}"
    )
except SyntaxError:
    mergingDataSESpecial = {}
mergingRemoveInputsFlag = "{{MergeRemoveFlag#PROD-2:Merging: remove input data flag True/False#True}}"
mergeCompressionLvl = "{{mergeCompressionLvl#PROD-2: compression level#HIGH}}"
mergeOutputVisFlag = "{{mergeOutputVisFlag#PROD-2: Visibility flag of output files#Y}}"
try:
    mergeOutputVisFlagSpecial = ast.literal_eval(
        "{{mergeOutputVisFlagSpecial#PROD-2: Special Visibility flag of output files (dict FType:Y|N)#}}"
    )
except SyntaxError:
    mergeOutputVisFlagSpecial = {}

pr.requestID = "{{ID}}"
pr.prodGroup = "{{inProPass}}" + "/" + "{{pDsc}}"
# used in case of a test e.g. certification etc.
pr.configName = "{{configName}}"
pr.configVersion = "{{configVersion}}"
# Other parameters from the request page
pr.dqFlag = "{{inDataQualityFlag}}"  # UNCHECKED
pr.dataTakingConditions = "{{simDesc}}"
pr.processingPass = "{{inProPass}}"
pr.bkFileType = "{{inFileType}}"
pr.eventType = "{{eventType}}"
pr.visibility = "Yes"
targetSite = "ALL"
sprucMulticoreFlag = mergeMulticoreFlag = "True"
sprucIDPolicy = mergingIDPolicy = "download"


w1 = ast.literal_eval(w1)
w2 = ast.literal_eval(w2)

validationFlag = ast.literal_eval(validationFlag)

mergeRemoveInputsFlag = ast.literal_eval(mergingRemoveInputsFlag)

if not w1 and not w2:
    gLogger.error("I told you to select at least one workflow!")
    DIRACexit(2)

pr.outConfigName = pr.configName

# In case we want just to test, we publish in the certification/test part of the BKK
if gConfig.getValue("DIRAC/Setup") == "LHCb-Certification":
    pr.outConfigName = "certification"
    if "lhcb" not in pr.configName.lower():  # in case of reconstruction
        pr.configVersion = "test"

if validationFlag:
    pr.outConfigName = "validation"

if w1:
    pr.prodsTypeList = ["Sprucing"]
    pr.outputSEs = [sprucDataSE]
    pr.specialOutputSEs = [sprucDataSESpecial]
    pr.stepsInProds = [list(range(1, len(pr.stepsList) + 1))]
    pr.removeInputsFlags = [False]
    pr.priorities = [sprucPriority]
    pr.cpus = [sprucCPU]
    pr.groupSizes = [sprucFilesPerJob]
    pr.plugins = [sprucPlugin]
    pr.inputs = [[]]
    pr.inputDataPolicies = [sprucIDPolicy]
    pr.bkQueries = ["Full"]
    pr.targets = [targetSite]
    pr.multicore = [sprucMulticoreFlag]
    pr.outputModes = ["Run"]
    pr.ancestorDepths = [sprucAncestorDepth]
    pr.compressionLvl = [sprucCompressionLvl] * len(pr.stepsInProds[0])
    pr.outputVisFlag = [{str(i + 1): sprucOutputVisFlag} for i in range(len(pr.stepsInProds[0]))]
    pr.specialOutputVisFlag = [{"1": sprucOutputVisFlagSpecial}]

elif w2:
    pr.prodsTypeList = ["Sprucing", "Merge"]
    pr.outputSEs = [sprucDataSE, mergingDataSE]
    pr.specialOutputSEs = [sprucDataSESpecial, mergingDataSESpecial]
    pr.stepsInProds = [list(range(1, len(pr.stepsList))), [len(pr.stepsList)]]
    pr.removeInputsFlags = [False, mergeRemoveInputsFlag]
    pr.priorities = [sprucPriority, mergingPriority]
    pr.cpus = [sprucCPU, mergingCPU]
    pr.groupSizes = [sprucFilesPerJob, mergingGroupSize]
    pr.plugins = [sprucPlugin, mergingPlugin]
    pr.inputs = [[], []]
    pr.inputDataPolicies = [sprucIDPolicy, mergingIDPolicy]
    pr.bkQueries = ["Full", "fromPreviousProd"]
    pr.targets = [targetSite, targetSite]
    pr.multicore = [sprucMulticoreFlag, mergeMulticoreFlag]
    pr.outputModes = ["Run", "Run"]
    pr.ancestorDepths = [sprucAncestorDepth, 0]
    pr.compressionLvl = [sprucCompressionLvl] * len(pr.stepsInProds[0]) + [mergeCompressionLvl] * len(
        pr.stepsInProds[1]
    )
    pr.outputVisFlag = [{"1": sprucOutputVisFlag}, {"2": mergeOutputVisFlag}]
    pr.specialOutputVisFlag = [{"1": sprucOutputVisFlagSpecial}, {"2": mergeOutputVisFlagSpecial}]


if not (pr.startRun and pr.endRun) and not pr.runsList:
    gLogger.error("Please do select a start and an end run, or a runs list")
    DIRACexit(2)

pr.buildAndLaunchRequest()



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Client for accessing data about Analysis Productions

Overview of concepts
--------------------

The basic objects stored for Analysis Productions are::

                  ┌─────────────┐
                  │ Publication │
                  └──────┬──────┘
                         *
                         │
                         1
  ┌─────────┐        ┌───┴────┐        ┌──────┐
  │ Request ├1──────*┤ Sample ├1──────*┤ User │
  └────┬────┘        └───┬────┘        └──────┘
       1                 1
       │                 │
       *                 *
  ┌────┴────┐         ┌──┴──┐
  │ AutoTag │         │ Tag │
  └─────────┘         └─────┘

* **Request**:
    A *Request* in the :py:mod:`.AnalysisProductionsDB` corresponds request in
    the :py:mod:`.ProductionRequestDB` with the type ``AnalysisProduction``.
    Each processed *Request* corresponds to 2 or more *Transformations* from
    the :py:mod:`.TransformationSystem` and the output LFNs are found by
    querying the Bookkeeping.
* **Sample**:
    A *Sample* wraps around a *Request* to allow a single sample to be reused.
    It can be uniquely identified by a ``WG``, ``analysis``, ``name`` and
    ``version`` where ``name`` comes from the original ``info.yaml`` file that
    was used to submit the production and ``version`` is the version of the tag
    in the ``AnalysisProductions`` data-package.
* **Tag**:
    A *Tag* is a `key` and `value` that can be used to query for a specific
    *Sample*. These can be manually defined by a _User_ to apply analysis specific
    metadata to a sample.
* **AutoTag**:
    Similar to a *Tag* how ever these are automatically computed from the
    *Bookkeeping* and applies to the underlying *Request* object. The keys of
    *AutoTags* are reserved and cannot be used by *Tags*.
* **User**:
    The nickname of the person who is able to modify and *Archive* a specific *Sample*.
* **Publication**:
    Denote that this was used for an LHCb result and the data for the underlying
    *Request* should be permanently preserved.

Archiving samples
-----------------

Arching a *Sample* is a way of denoting that is no longer useful. Typically
because it has been superseded be another *Sample*. When all child *Samples*
of a *Request* have been archived it is an indication that the data of the
*Request* can be permanently deleted.

Time travel
-----------

Most methods have an optional ``at_time`` argument which can be used to query
the database at a specific point in the past. This allows for the database to
continue to be chaged over time and while still allowing for high quality
analysis preservation.

Populating the database
-----------------------

The database is populated automatically by the :py:mod:`.APSyncAgent`.

Legacy considerations
---------------------

Prior to the creation of the ``AnalysisProductions`` data-package, the
``WG/CharmConfig`` and ``WG/CharmWGProd`` were used to automatically submit
production requests. Samples from this system were manually imported and are
mostly the same as those submitted by ``AnalysisProductions`` except the
``version`` of the production is of the form ``vXrYpZ (PackageName)``.

API Reference
-------------
"""
from typing import Any, Optional as Opt
from datetime import datetime

from DIRAC.Core.Base.Client import Client, createClient


@createClient("ProductionManagement/AnalysisProductions")
class AnalysisProductionsClient(Client):
    """Provides access to the data stored in the Analysis Productions database.

    This class wraps all the RPC calls to default parameters to be injected.
    """

    def __init__(self, url=None, **kwargs):
        super().__init__(**kwargs)
        self.setServer("ProductionManagement/AnalysisProductions")
        if url:
            self.setServer(url)

    def listAnalyses(self, *, at_time: Opt[datetime] = None):
        """Return the mapping of known WGs to analyses

        :param at_time: The datetime at which this query should be ran, defaults to now
        :returns: ``dict`` mapping the WG name to a list of analyses
        """
        return self.executeRPC(at_time, call="listAnalyses")

    def listAnalyses2(self, *, at_time: Opt[datetime] = None):
        """Return the mapping of known WGs to analyses

        :param at_time: The datetime at which this query should be ran, defaults to now
        :returns: ``list`` with each entry being a ``dict``  of high-level information about each analysis
        """
        return self.executeRPC(at_time, call="listAnalyses2")

    def listRequests(self):
        """Return the mapping of known WGs to analyses

        :returns: ``list`` with each entry being a ``dict``  of high-level information about each sample
        """
        return self.executeRPC(call="listRequests")

    def getOwners(self, *, wg: str, analysis: str):
        """Return the list of owners for a specific analysis

        :returns: ``dict`` mapping the WG name to a list of analyses
        """
        return self.executeRPC(wg, analysis, call="getOwners")

    def setOwners(self, *, wg: str, analysis: str, owners: list[str]):
        """Set the list of owners for a specific analysis

        :returns: ``dict`` mapping the WG name to a list of analyses
        """
        return self.executeRPC(wg, analysis, owners, call="setOwners")

    def getProductions(
        self,
        *,
        wg: Opt[str] = None,
        analysis: Opt[str] = None,
        version: Opt[str] = None,
        name: Opt[str] = None,
        state: Opt[str] = None,
        with_lfns: bool = True,
        with_pfns: bool = True,
        with_transformations: bool = False,
        at_time: Opt[datetime] = None,
        show_archived: bool = False,
        require_has_publication: bool = False,
    ):
        """Return the list of productions for a specific analysis

        :param wg: The WG to which the analysis belongs
        :param analysis: The name of the analysis
        :param state: The state of the underlying request
        :param version: The version of the data package to include
        :param state: The name of the sample
        :param with_lfns: Include LFNs in the output, default True
        :param with_pfns: Include PFNs in the output, default True
        :param with_transformations: Include transformation info in the output, default False
        :param at_time: The datetime at which this query should be ran, defaults to now
        :returns: A dictionary of name -> version -> requestID
        """
        return self.executeRPC(
            wg,
            analysis,
            version,
            name,
            state,
            with_lfns,
            with_pfns,
            with_transformations,
            at_time,
            show_archived,
            require_has_publication,
            call="getProductions",
        )

    def getArchivedRequests(
        self,
        *,
        state: Opt[str] = None,
        with_lfns: bool = False,
        with_pfns: bool = False,
        with_transformations: bool = False,
    ):
        """Return the list of requests which are no longer used for analysis

        :param state: The state of the underlying request
        :param with_lfns: Include LFNs in the output, default False
        :param with_pfns: Include PFNs in the output, default False
        :param with_transformations: Include transformation info in the output, default False
        :returns: A dictionary of name -> version -> requestID
        """
        return self.executeRPC(state, with_lfns, with_pfns, with_transformations, call="getArchivedRequests")

    def getTags(self, *, wg: Opt[str] = None, analysis: Opt[str] = None, at_time: Opt[datetime] = None):
        """Return the list of productions for a specific analysis

        :param wg: The WG to which the analysis belongs
        :param analysis: The name of the analysis
        :param at_time: The datetime at which this query should be ran, defaults to now
        :returns: A dictionary of name -> version -> requestID
        """
        return self.executeRPC(wg, analysis, at_time, call="getTags")

    def getKnownAutoTags(self):
        """Return a list of all known automatic tag names"""
        return self.executeRPC(call="getKnownAutoTags")

    def setState(self, newState: dict[int, dict[str, Any]]):
        """Update the state of requests in the Analysis Productions database

        :param newState: Dictionary of the form {requestID: {"state" : "newState", "progress": 0.5}}}
        :returns: S_OK() || S_ERROR()
        """
        return self.executeRPC(newState, call="setState")

    def registerTransformations(self, transformations: dict[int, list[dict]]):
        """Update the state of requests in the Analysis Productions database

        :param transformations: Dictionary of the form {requestID: [tInfo, tInfo, ...]}}
        :returns: S_OK() || S_ERROR()
        """
        return self.executeRPC(transformations, call="registerTransformations")

    def deregisterTransformations(self, tIDs: dict[int, list[int]]):
        """Deregister the some transformations associated with a request

        :param tIDs: Dictionary of the form {requestID: [tID, tID, ...]}}
        :returns: S_OK() || S_ERROR()
        """
        return self.executeRPC(tIDs, call="deregisterTransformations")

    def registerRequests(self, requests):
        """Add a new production to the database

        :param requests: List of dictionaries
        :returns: S_OK() || S_ERROR()
        """
        return self.executeRPC(requests, call="registerRequests")

    def addRequestsToAnalysis(self, wg: str, analysis: str, requests: list[tuple[int, str]]):
        """Add a new production to the database

        :param wg: The WG to which the analysis belongs
        :param analysis: The analysis which should have the samples added to
        :param requests: Dict of request IDs to filetypes
        :returns: S_OK() || S_ERROR()
        """
        return self.executeRPC(wg, analysis, requests, call="addRequestsToAnalysis")

    def archiveSamples(self, samples):
        """Archive an analysis Sample

        :param samples: List of sample IDs
        :returns: S_OK() || S_ERROR()
        """
        return self.executeRPC(samples, call="archiveSamples")

    def archiveSamplesAtSpecificTime(self, samples, archive_time):
        """Archive an analysis Sample at the given time

        :param samples: List of sample IDs
        :param archive_time: datetime object of when to archive the sample
        :returns: S_OK() || S_ERROR()
        """
        return self.executeRPC(samples, archive_time, call="archiveSamplesAtSpecificTime")

    def delayHousekeepingInteractionDue(self, samples, next_interaction_due):
        """Set the next due date for housekeeping notifications.

        :param samples: List of sample IDs
        :param next_interaction_due: datetime object for when next to send housekeeping notifications.
        :returns: S_OK() || S_ERROR()
        """
        return self.executeRPC(samples, next_interaction_due, call="delayHousekeepingInteractionDue")

    def getHousekeepingInteractionDueNow(self):
        """Return requests that are due for housekeeping notifications now.

        :returns: S_OK() || S_ERROR()
        """
        return self.executeRPC(call="getHousekeepingInteractionDueNow")

    def addPublication(self, samples, number):
        """Add a publication number to the specified samples.

        :param samples: List of sample IDs
        :param number: publication number as a string (<64chars).
        :returns: S_OK() || S_ERROR()
        """
        return self.executeRPC(samples, number, call="addPublication")

    def getPublications(self, sample_ids: Opt[list] = None):
        """Get publication numbers for the specified sample.

        :param samples: List of sample IDs
        :returns: S_OK() || S_ERROR()
        """
        return self.executeRPC(sample_ids, call="getPublications")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: StorageUsageClient

.. module: StorageUsageClient

:synopsis: Lightweight possbile client to the StorageUsageDB.
"""

from DIRAC.Core.Base.Client import Client, createClient


@createClient("DataManagement/StorageUsage")
class StorageUsageClient(Client):
    """
    .. class:: StorageUsageClient
    """

    def __init__(self, url=None, **kwargs):
        """c'tor."""
        super().__init__(**kwargs)
        if url:
            self.setServer(url)
        self.setServer("DataManagement/StorageUsage")
